# -*- coding: utf-8 -*-
"""
Created on Sun May 12 16:34:44 2019

@author: Alexandre
"""

###############################################################################
import numpy as np
###############################################################################
from pyro.control  import robotcontrollers
from pyro.dynamic  import manipulator
###############################################################################

torque_controlled_robot      = manipulator.TwoLinkManipulator()

# Target
q_desired = np.array([0.5,0.5])
r_desired = torque_controlled_robot.forward_kinematic_effector( q_desired )

# effector PID

dof = 2

effector_pid      = robotcontrollers.EndEffectorPID( torque_controlled_robot )
effector_pid.rbar = r_desired
effector_pid.kp   = np.array([100, 100 ])
effector_pid.kd   = np.array([  0,   0 ])
effector_pid.ki   = np.array([ 50,  50 ])


# Closed-loops

robot_with_effector_pid    = effector_pid + torque_controlled_robot 

# Simulations
tf = 20
robot_with_effector_pid.x0 = np.array([0,0,0,0,0,0])
robot_with_effector_pid.compute_trajectory( tf )
robot_with_effector_pid.plot_trajectory('xu')
robot_with_effector_pid.plot_internal_controller_states()
robot_with_effector_pid.animate_simulation()



# -*- coding: utf-8 -*-
"""
Created on 19/11/2019

@author: Pierre
"""
###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic import pendulum
from adaptive_computed_torque import SinglePendulumAdaptativeController
###############################################################################

sys = pendulum.SinglePendulum()


sys.m1       = 1
sys.l1       = 2
sys.lc1      = 1
sys.I1       = 0
sys.gravity  = 9.81 


# Controller

ctl = SinglePendulumAdaptativeController(sys)

# Sys ID initial guess
m1_hat = 0
l1_hat = 0

ctl.z0[0] = m1_hat * l1_hat ** 2
ctl.z0[1] = m1_hat * l1_hat * 9.81

def t2traj( t , r ):
    """  """
    a = 1.5
    w = 2.5
    v = 1.0
    
    ddq_d          =   - a * w * w * np.sin( w * t)
    dq_d           =   a * w * np.cos( w * t) + v
    q_d            =   r + a * np.sin( w * t) + v * t
    
    return [ ddq_d , dq_d , q_d ] 



# Controller Param
ctl.K      = 10
ctl.lam    = 2
ctl.P[0,0] = 10
ctl.P[1,1] = 30
ctl.get_traj = t2traj


# Set Point
q_target = np.array([0.0]) # np.array([3.14])
ctl.rbar = q_target

# New cl-dynamic
cl_sys = ctl + sys

cl_sys.state_label[2] = 'H'
cl_sys.state_label[3] = 'mgl'


# Simultation
cl_sys.x0[0]  = 0

cl_sys.compute_trajectory(tf=30, n=20001) # solver='euler')
cl_sys.plot_phase_plane_trajectory()
cl_sys.plot_trajectory_with_internal_states()
cl_sys.animate_simulation()


# -*- coding: utf-8 -*-
"""
Created on Sun May 12 16:34:44 2019

@author: Alexandre
"""

###############################################################################
import numpy as np
###############################################################################
from pyro.control  import robotcontrollers
from pyro.dynamic  import manipulator
###############################################################################

torque_controlled_robot = manipulator.ThreeLinkManipulator3D()
speed_controlled_robot  = manipulator.SpeedControlledManipulator.from_manipulator( 
                                             torque_controlled_robot )


kinematic_controller = robotcontrollers.EndEffectorKinematicController( speed_controlled_robot )
kinematic_controller.rbar = np.array([1.0,1.0,1.0])
    
closed_loop_robot = kinematic_controller + speed_controlled_robot
    
closed_loop_robot.x0 = np.array([0.1,0.1,0.1])
closed_loop_robot.plot_trajectory('xu')
closed_loop_robot.animate_simulation( time_factor_video=2, is_3d=True )


###############################################################################
import numpy as np
import matplotlib.pyplot as plt
###############################################################################
from asimov import Asimov
from pyro.control import robotcontrollers
###############################################################################
kp = 40
ki = 0
kd = 0

asimov = Asimov()  # Asimov
asimov.x0 = np.array([-np.pi/4, -3*np.pi/4, np.pi/2, 0, 0, 0])  # Position initiale

qd = np.array([0.5, -np.pi/4, 0.5])  # Cible au joint

ctl = robotcontrollers.JointPID(3, kp, ki, kd)  # Déclaration du controlleur
ctl.rbar = qd
ctl.kd = np.array([5, 8, 1])

closed_loop_robot = ctl + asimov  # Système boucle fermé

closed_loop_robot.plot_trajectory()  # Calcul de la trajectoire
closed_loop_robot.animate_simulation( is_3d=True )


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 30 14:04:51 2021

@author: alex
"""

import numpy as np

from pyro.dynamic.pendulum                import Acrobot
from pyro.planning.trajectoryoptimisation import DirectCollocationTrajectoryOptimisation
from pyro.planning.randomtree             import RRT


sys = Acrobot()

#Max/Min torque
sys.u_ub[0] = +20
sys.u_lb[0] = -20

# RRT

x_start = np.array([-3.14,0,0,0])
x_goal  = np.array([0,0,0,0])

rrt = RRT( sys )
    
#rrt.u_options = [ sys.u_ub , sys.u_lb ]
t = 15.0

rrt.u_options = [
        np.array([+t]),
        np.array([-t]),
        np.array([0.0])
        ]

rrt.x_start = x_start
rrt.x_goal  = x_goal

rrt.goal_radius          = 1.5
rrt.dt                   = 0.1
rrt.max_nodes            = 2000
rrt.max_solution_time    = 10.0
rrt.max_distance_compute = 1000
rrt.dyna_plot            = False

rrt.find_path_to_goal()

tf = rrt.traj.t[-1]
n  = 40
dt = tf / n

planner = DirectCollocationTrajectoryOptimisation( sys , dt , n )

planner.x_start = x_start
planner.x_goal  = x_goal

planner.set_initial_trajectory_guest( rrt.traj )

planner.maxiter = 500
planner.compute_optimal_trajectory()
planner.show_solution()
planner.animate_solution()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu May 14 23:19:16 2020

@author: alex
------------------------------------


Fichier d'amorce pour les livrables de la problématique GRO640'


"""

import numpy as np

from pyro.control  import robotcontrollers
from pyro.control.robotcontrollers import EndEffectorPD
from pyro.control.robotcontrollers import EndEffectorKinematicController


###################
# Part 1
###################

def dh2T( r , d , theta, alpha ):
    """

    Parameters
    ----------
    r     : float 1x1
    d     : float 1x1
    theta : float 1x1
    alpha : float 1x1
    
    4 paramètres de DH

    Returns
    -------
    T     : float 4x4 (numpy array)
            Matrice de transformation

    """
    
    T = np.zeros((4,4))
    
    ###################
    # Votre code ici
    ###################
    
    return T



def dhs2T( r , d , theta, alpha ):
    """

    Parameters
    ----------
    r     : float nx1
    d     : float nx1
    theta : float nx1
    alpha : float nx1
    
    Colonnes de paramètre de DH

    Returns
    -------
    WTT     : float 4x4 (numpy array)
              Matrice de transformation totale de l'outil

    """
    
    WTT = np.zeros((4,4))
    
    ###################
    # Votre code ici
    ###################
    
    return WTT


def f(q):
    """
    

    Parameters
    ----------
    q : float 6x1
        Joint space coordinates

    Returns
    -------
    r : float 3x1 
        Effector (x,y,z) position

    """
    r = np.zeros((3,1))
    
    ###################
    # Votre code ici
    ###################
    
    return r


###################
# Part 2
###################
    
class CustomPositionController( EndEffectorKinematicController ) :
    
    ############################
    def __init__(self, manipulator ):
        """ """
        
        EndEffectorKinematicController.__init__( self, manipulator, 1)
        
        ###################################################
        # Vos paramètres de loi de commande ici !!
        ###################################################
        
    
    #############################
    def c( self , y , r , t = 0 ):
        """ 
        Feedback law: u = c(y,r,t)
        
        INPUTS
        y = q   : sensor signal vector  = joint angular positions      dof x 1
        r = r_d : reference signal vector  = desired effector position   e x 1
        t       : time                                                   1 x 1
        
        OUPUTS
        u = dq  : control inputs vector =  joint velocities             dof x 1
        
        """
        
        # Feedback from sensors
        q = y
        
        # Jacobian computation
        J = self.J( q )
        
        # Ref
        r_desired   = r
        r_actual    = self.fwd_kin( q )
        
        # Error
        e  = r_desired - r_actual
        
        ################
        dq = np.zeros( self.m )  # place-holder de bonne dimension
        
        ##################################
        # Votre loi de commande ici !!!
        ##################################

        
        return dq
    
    
###################
# Part 3
###################
        

        
class CustomDrillingController( robotcontrollers.RobotController ) :
    """ 

    """
    
    ############################
    def __init__(self, robot_model ):
        """ """
        
        super().__init__( dof = 3 )
        
        self.robot_model = robot_model
        
        # Label
        self.name = 'Custom Drilling Controller'
        
        
    #############################
    def c( self , y , r , t = 0 ):
        """ 
        Feedback static computation u = c(y,r,t)
        
        INPUTS
        y  : sensor signal vector     p x 1
        r  : reference signal vector  k x 1
        t  : time                     1 x 1
        
        OUPUTS
        u  : control inputs vector    m x 1
        
        """
        
        # Ref
        f_e = r
        
        # Feedback from sensors
        x = y
        [ q , dq ] = self.x2q( x )
        
        # Robot model
        r = self.robot_model.forward_kinematic_effector( q ) # End-effector actual position
        J = self.robot_model.J( q )      # Jacobian matrix
        g = self.robot_model.g( q )      # Gravity vector
        H = self.robot_model.H( q )      # Inertia matrix
        C = self.robot_model.C( q , dq ) # Coriolis matrix
            
        ##################################
        # Votre loi de commande ici !!!
        ##################################
        
        u = np.zeros(self.m)  # place-holder de bonne dimension
        
        return u
        
    
###################
# Part 4
###################
        
    
def goal2r( r_0 , r_f , t_f ):
    """
    
    Parameters
    ----------
    r_0 : numpy array float 3 x 1
        effector initial position
    r_f : numpy array float 3 x 1
        effector final position
    t_f : float
        time 

    Returns
    -------
    r   : numpy array float 3 x l
    dr  : numpy array float 3 x l
    ddr : numpy array float 3 x l

    """
    # Time discretization
    l = 1000 # nb of time steps
    
    # Number of DoF for the effector only
    m = 3
    
    r = np.zeros((m,l))
    dr = np.zeros((m,l))
    ddr = np.zeros((m,l))
    
    #################################
    # Votre code ici !!!
    ##################################
    
    
    return r, dr, ddr


def r2q( r, dr, ddr , manipulator ):
    """

    Parameters
    ----------
    r   : numpy array float 3 x l
    dr  : numpy array float 3 x l
    ddr : numpy array float 3 x l
    
    manipulator : pyro object 

    Returns
    -------
    q   : numpy array float 3 x l
    dq  : numpy array float 3 x l
    ddq : numpy array float 3 x l

    """
    # Time discretization
    l = r.shape[1]
    
    # Number of DoF
    n = 3
    
    # Output dimensions
    q = np.zeros((n,l))
    dq = np.zeros((n,l))
    ddq = np.zeros((n,l))
    
    #################################
    # Votre code ici !!!
    ##################################
    
    
    return q, dq, ddq



def q2torque( q, dq, ddq , manipulator ):
    """

    Parameters
    ----------
    q   : numpy array float 3 x l
    dq  : numpy array float 3 x l
    ddq : numpy array float 3 x l
    
    manipulator : pyro object 

    Returns
    -------
    tau   : numpy array float 3 x l

    """
    # Time discretization
    l = q.shape[1]
    
    # Number of DoF
    n = 3
    
    # Output dimensions
    tau = np.zeros((n,l))
    
    #################################
    # Votre code ici !!!
    ##################################
    
    
    return tau


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Oct  3 08:27:06 2021

@author: alex
"""

import numpy as np

from pyro.dynamic.drone                   import Drone2D
from pyro.planning.trajectoryoptimisation import DirectCollocationTrajectoryOptimisation
from pyro.dynamic.statespace              import linearize

from pyro.analysis.costfunction           import QuadraticCostFunctionVectorized


# Non-linear model
sys = Drone2D()

sys.xbar =  np.array([0,0,0.0,0,0,0])
sys.ubar =  np.array([0.5,0.5]) * sys.mass * sys.gravity
    
# Linear model
ss  = linearize( sys , 0.01 )

cf = QuadraticCostFunctionVectorized( sys.n, sys.m )

planner = DirectCollocationTrajectoryOptimisation( ss , cost_function = cf )

planner.x_start = np.array([-5,0,0,0,0,0])
planner.x_goal  = np.array([0,0,0,0,0,0])

planner.compute_optimal_trajectory()
planner.show_solution()
planner.animate_solution()





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Oct 16 22:27:47 2022

@author: alex
"""

import numpy as np

from pyro.dynamic  import suspension

sys  = suspension.QuarterCarOnRoughTerrain()

sys.k  = 10.0
sys.vx = 10.0


# Simulation and animation
sys.compute_trajectory( 10, 10001, 'euler')
sys.plot_trajectory('xu')
sys.animate_simulation()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu May  4 10:01:30 2023

@author: alex
"""

###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic   import cartpole
from pyro.analysis  import costfunction
from pyro.planning  import trajectoryoptimisation
from pyro.control   import lqr
###############################################################################

##############
# System
##############

sys  = cartpole.CartPole()

sys.u_ub[0] = +20
sys.u_lb[0] = -20

################
# Cost function
################

cf  = costfunction.QuadraticCostFunction.from_sys( sys )

cf.Q[0,0] = 1.0
cf.Q[1,1] = 1.0
cf.R[0,0] = 1.0

###########################
# Trajectory optimization
##########################

planner = trajectoryoptimisation.DirectCollocationTrajectoryOptimisation( 
            sys , 
            dt = 0.1 , 
            grid = 40 ,
            cost_function = cf 
            )


planner.x_start = np.array([0,0,0,0])
planner.x_goal  = np.array([0,np.pi,0,0])

planner.init_dynamic_plot()
planner.maxiter = 500
planner.compute_optimal_trajectory()

traj = planner.traj


###########################
# LQR Controller
##########################

cf.R[0,0] = 5.0

ctl = lqr.TrajectoryLQRController( sys , traj , cf )


###########################
# Simulation
##########################

cl_sys = ctl + sys

cl_sys.x0[0] = 0.5
cl_sys.x0[1] = 0.5
    
cl_sys.compute_trajectory( tf = 10.0 )
cl_sys.plot_trajectory('xu')
cl_sys.animate_simulation( time_factor_video=1.0 )




# -*- coding: utf-8 -*-
"""
Created on Fri Nov 16 12:01:07 2018

@author: Alexandre
"""

###############################################################################
import numpy as np
import matplotlib.pyplot as plt
###############################################################################
import advanced_vehicles
import test_vehicle_controllers
###############################################################################

# "Fake" controller - Varying inputs (delta, F_xf, F_xr) throughout time (change in linear.py)
ctl = test_vehicle_controllers.dynLongForcesInputs()
# Vehicule dynamical system
sys = advanced_vehicles.LateralDynamicBicycleModelwithForceInputs()
# Add the inputs to the dynamical system
cl_sys = ctl + sys
# Plot open-loop behavior (ex: np.array[intial_conditions], time_of_simulation)
cl_sys.x0 = np.array([0,0,0,0,0,0])
cl_sys.compute_trajectory( 48 )
cl_sys.plot_trajectory('x')

# Rebuild x,u and t from simulation
x = cl_sys.traj.x
u = cl_sys.traj.u
t = cl_sys.traj.t 

# Compute tire forces and slip angles for graphical purpose  
F_nf = sys.mass*sys.g*sys.b/(sys.b+sys.a)
F_nr = sys.mass*sys.g*sys.a/(sys.b+sys.a)
mu = sys.mu
max_F_f = F_nf*mu
max_F_r = F_nr*mu
max_alpha_stat = sys.max_alpha_stat
slip_ratio_f = max_F_f/(max_alpha_stat) #0.12 is the max slipping angle in rad before kinetic slip
slip_ratio_r = max_F_r/(max_alpha_stat) 
# Init. vectors for display purposes
F_yf = np.zeros(len(t))
F_yr = np.zeros(len(t))
slip_f = np.zeros(len(t))
slip_r = np.zeros(len(t))
maxF_f = np.zeros(len(t))
maxF_r = np.zeros(len(t))
# Iterate through time to compute slipping angles and lateral forces
for i in range(len(t)-1):
    if (x[i,0] == 0):
        slip_f[i] = 0
        slip_r[i] = 0
        F_yf[i]   = 0
        F_yr[i]   = 0
    else:
        slip_f[i] = np.arctan((x[i,1]+sys.a*x[i,2])/x[i,0])-u[i,0]
        slip_r[i] = np.arctan((x[i,1]-sys.b*x[i,2])/x[i,0])
        if (slip_f[i]<-0.12):
            F_yf[i] = max_F_f
        elif (slip_f[i] > 0.12):
            F_yf[i] = -max_F_f
        else:
            F_yf[i] = -slip_ratio_f*slip_f[i]
            
        if (slip_r[i]<-0.12):
            F_yr[i] = max_F_r
        elif (slip_r[i] > 0.12):
            F_yr[i] = -max_F_r
        else:
            F_yr[i] = -slip_ratio_r*slip_r[i]
    maxF_f[i] = max_F_f
    maxF_r[i] = max_F_r

# Plot forces
figsize   = (7, 4)
dpi       = 100
plt.figure(2, figsize=figsize, dpi=dpi)
plt.title('Lateral forces for the lateral dynamic\n model with steering angle and longitudinal forces as inputs', fontsize=20)
plt.plot(t[:-1], F_yf[:-1], label = 'F_yf')
plt.plot(t[:-1], F_yr[:-1], label = 'F_yr')
plt.plot(t[:-1], maxF_f[:-1],'--', label = 'Max front force')
plt.plot(t[:-1], maxF_r[:-1],'--', label = 'Max rear force')
plt.legend(fontsize = '15')
plt.xlabel('Temps (s)')
plt.ylabel('Force (N)')
plt.show()


# Plot trajectory of the vehicle's CG
slip = 1
t_start_slip = 'No slip' #No slip
plt.figure(3,figsize=figsize, dpi=dpi)
for i in range(len(t)-1):
    if (F_yf[i]==max_F_f or F_yr[i]==max_F_r):
        if (slip == 1):            
            t_start_slip = i
            slip = 0
        else:
            t_stop_slip = i
    else:
        pass
if (t_start_slip == 'No slip'):
    plt.plot(x[:,4],x[:,5],'-b', label = 'No slip')     
else:
    plt.plot((x[t_start_slip:t_stop_slip,4]),(x[t_start_slip:t_stop_slip,5]),'-r',label='Slip')
    plt.plot((x[0:t_start_slip,4]),(x[0:t_start_slip,5]),'-b', label='No slip')
    plt.plot((x[t_stop_slip:len(t)-1,4]),(x[t_stop_slip:len(t)-1,5]),'-b')    
plt.legend(fontsize ='15')
plt.show()

# Animate the simulation
cl_sys.plot_trajectory('x')
cl_sys.plot_trajectory('u')
cl_sys.animate_simulation()



# -*- coding: utf-8 -*-
"""
Created on Mon Nov 12 20:28:17 2018

@author: Alexandre
"""
##############################################################################
import numpy as np
##############################################################################


##############################################################################
# Dynamics
##############################################################################

from pyro.dynamic  import pendulum

# Define the dynamical system to control
sys  = pendulum.SinglePendulum()

sys.plot_phase_plane()

#Max/Min torque
sys.u_ub[0] = +5
sys.u_lb[0] = -5

sys.xbar  = np.array([ -3.14 , 0 ]) # target state [ position , velocity]
sys.x0    = np.array([ -0.00 , 0 ]) # initial state

##############################################################################
# Cost Function
##############################################################################

from pyro.analysis import costfunction

cf = costfunction.QuadraticCostFunction.from_sys( sys ) 

cf.INF  = 300    # The value iteration algo needs this parameter

cf.Q[0,0] = 1
cf.Q[1,1] = 0.1
cf.R[0,0] = 1

print('Q=\n',cf.Q)
print('R=\n',cf.R)

sys.cost_function = cf

##############################################################################
# Sub-optimal PID
##############################################################################

from pyro.control  import linear

pid      = linear.ProportionalController(1,2) # 1 output from 2 inputs
pid.rbar = sys.xbar                           # target
pid.K    = np.array([10,4])                   # gain matrix

print('PID K=\n',pid.K)
pid.plot_control_law( sys = sys )

sys_with_pid = pid + sys
sys_with_pid.plot_trajectory('xuj')
sys_with_pid.animate_simulation()

print('Trajectory cost: ', sys_with_pid.traj.J[-1])

##############################################################################
# LQR
##############################################################################

from pyro.dynamic  import statespace
from pyro.control  import lqr

# Linear model
ss  = statespace.linearize( sys )

print('A=\n',ss.A)
print('B=\n',ss.B)

lqr_ctl = lqr.synthesize_lqr_controller( ss , cf , sys.xbar)

print('LQR K=\n',lqr_ctl.K)
lqr_ctl.plot_control_law( sys = sys )

sys_with_lqr = lqr_ctl + sys
sys_with_lqr.plot_trajectory('xuj')
sys_with_lqr.animate_simulation()

print('Trajectory cost: ', sys_with_lqr.traj.J[-1])

##############################################################################
# VI
##############################################################################

from pyro.planning import dynamicprogramming 
from pyro.planning import discretizer

grid_sys = discretizer.GridDynamicSystem( sys , [101,101] , [11] )

dp = dynamicprogramming.DynamicProgrammingWithLookUpTable( grid_sys, cf)

dp.solve_bellman_equation()

vi_ctl = dp.get_lookup_table_controller()

vi_ctl.plot_control_law( sys = sys , n=1000)

sys_with_vi = vi_ctl + sys
sys_with_vi.plot_trajectory('xuj')
sys_with_vi.animate_simulation()

print('Trajectory cost: ', sys_with_vi.traj.J[-1])

##############################################################################
# Direct Collocation Trajectory Optimisation
##############################################################################

from pyro.planning import trajectoryoptimisation

planner = trajectoryoptimisation.DirectCollocationTrajectoryOptimisation( sys ,
                                                                         dt = 0.1,
                                                                         grid = 50)

planner.x_start = sys.x0
planner.x_goal  = sys.xbar

planner.compute_optimal_trajectory()
planner.show_solution()

##############################################################################
# Stabilizing the optimal trajectory
##############################################################################

from pyro.control  import nonlinear

traj_ctl  = nonlinear.ComputedTorqueController( sys , planner.traj)  

traj_ctl.w0   = 2
traj_ctl.zeta = 1
traj_ctl.rbar = sys.xbar[0:1]

traj_ctl.plot_control_law( sys = sys , n=100)


sys_with_traj_ctl = traj_ctl + sys
sys_with_traj_ctl.plot_trajectory('xuj')
sys_with_traj_ctl.animate_simulation()

print('Trajectory cost: ', sys_with_traj_ctl.traj.J[-1])




# -*- coding: utf-8 -*-
"""
Created on Tue Oct 23 20:45:37 2018

@author: Alexandre
"""


###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic import system
###############################################################################

###############################################################################
        
class MechanicalSystem( system.ContinuousDynamicSystem ):
    """ 
    Mechanical system with Equation of Motion in the form of
    -------------------------------------------------------
    H(q) ddq + C(q,dq) dq + d(q,dq) + g(q) = B(q) u
    -------------------------------------------------------
    q      :  dim = (dof, 1)   : position variables 
    dq     :  dim = (dof, 1)   : velocity variables     
    ddq    :  dim = (dof, 1)   : acceleration variables
    u      :  dim = (m, 1)     : force input variables
    H(q)   :  dim = (dof, dof) : inertia matrix
    C(q)   :  dim = (dof, dof) : corriolis matrix
    B(q)   :  dim = (dof, m)   : actuator matrix
    ddq    :  dim = (dof, 1)   : acceleration variables
    d(q,dq):  dim = (dof, 1)   : state-dependent dissipative forces
    g(q)   :  dim = (dof, 1)   : state-dependent conservatives forces
    
    """
    
    ############################
    def __init__(self, dof = 1 , actuators = None):
        """ """
        
        # Degree of Freedom
        self.dof = dof
        
        # Nb of actuators
        if actuators == None:   # If not specifyied the sys is fully actuated
            actuators = dof
        
        # Dimensions
        n = dof * 2 
        m = actuators
        p = dof * 2
        
        # initialize standard params
        system.ContinuousDynamicSystem.__init__(self, n, m, p)
        
        # Name
        self.name = str(dof) + 'DoF Mechanical System'
        
        # Labels, bounds and units
        for i in range(dof):
            # joint angle states
            self.x_ub[i] = + np.pi * 2
            self.x_lb[i] = - np.pi * 2
            self.state_label[i] = 'Angle '+ str(i)
            self.state_units[i] = '[rad]'
            # joint velocity states
            self.x_ub[i+dof] = + np.pi * 2
            self.x_lb[i+dof] = - np.pi * 2
            self.state_label[i+dof] = 'Velocity ' + str(i)
            self.state_units[i+dof] = '[rad/sec]'
        for i in range(actuators):
            self.u_ub[i] = + 5
            self.u_lb[i] = - 5
            self.input_label[i] = 'Torque ' + str(i)
            self.input_units[i] ='[Nm]'
        self.output_label = self.state_label
        self.output_units = self.state_units
            
    ###########################################################################
    # The following functions needs to be overloaded by child classes
    # to represent the dynamic of the system
    ###########################################################################
    
    ###########################################################################
    def H(self, q ):
        """ 
        Inertia matrix 
        ----------------------------------
        dim( H ) = ( dof , dof )
        
        such that --> Kinetic Energy = 0.5 * dq^T * H(q) * dq
        
        """  
        
        H = np.diag( np.ones( self.dof ) ) # Default is identity matrix
        
        return H
    
    
    ###########################################################################
    def C(self, q , dq ):
        """ 
        Corriolis and Centrifugal Matrix 
        ------------------------------------
        dim( C ) = ( dof , dof )
        
        such that --> d H / dt =  C + C^T
        
        
        """ 
        
        C = np.zeros( ( self.dof , self.dof ) ) # Default is zeros matrix
        
        return C
    
    
    ###########################################################################
    def B(self, q ):
        """ 
        Actuator Matrix  : dof x m
        """
        
        B = np.zeros( ( self.dof , self.m ) )
        
        for i in range(min(self.m,self.dof)):
            B[i,i] = 1                # Diag matrix for the first m rows
        
        return B
    
    
    ###########################################################################
    def g(self, q ):
        """ 
        Gravitationnal forces vector : dof x 1
        """
        
        g = np.zeros( self.dof ) # Default is zero vector
        
        return g
    
        
    ###########################################################################
    def d(self, q , dq ):
        """ 
        State-dependent dissipative forces : dof x 1
        """
        
        d = np.zeros(self.dof ) # Default is zero vector
        
        return d
    
    
    ###########################################################################
    # No need to overwrite the following functions for custom system
    ###########################################################################
    
    #############################
    def x2q( self, x ):
        """ from state vector (x) to angle and speeds (q,dq) """
        
        q  = x[ 0        : self.dof ]
        dq = x[ self.dof : self.n   ]
        
        return [ q , dq ]
        
        
    #############################
    def q2x( self, q , dq ):
        """ from angle and speeds (q,dq) to state vector (x) """
        
        x = np.zeros( self.n )
        
        x[ 0        : self.dof ] = q
        x[ self.dof : self.n   ] = dq
        
        return x
    
    
    #############################
    def xut2q( self, x , u , t ):
        """ compute configuration variables """
        
        return self.x2q(x)[0]
    
    
    ##############################
    def generalized_forces(self, q  , dq  , ddq , t = 0 ):
        """ Computed generalized forces given a trajectory """  
        
        H = self.H( q )
        C = self.C( q , dq )
        g = self.g( q )
        d = self.d( q , dq )
                
        # Generalized forces
        forces = np.dot( H , ddq ) + np.dot( C , dq ) + g + d
        
        return forces
    
    
    ##############################
    def actuator_forces(self, q  , dq  , ddq , t = 0 ):
        """ Computed actuator forces given a trajectory (inverse dynamic) """  
        
        if self.dof == self.m:
        
            B = self.B( q )
                    
            # Generalized forces
            forces = self.generalized_forces( q , dq , ddq , t )
            
            # Actuator forces
            u = np.dot( np.linalg.inv( B ) , forces )
            
            return u
        
        else:
            
            raise NotImplementedError
        
    
    ##############################
    def ddq(self, q , dq , u , t = 0 ):
        """ Computed accelerations given actuator forces (foward dynamic) """  
        
        H = self.H( q )
        C = self.C( q , dq )
        g = self.g( q  )
        d = self.d( q , dq)
        B = self.B( q )
        
        ddq = np.dot( np.linalg.inv( H ) ,  ( np.dot( B , u )  
                                            - np.dot( C , dq ) - g - d ) )
        
        return ddq
    
    
    ###########################################################################
    def f(self, x , u , t = 0 ):
        """ 
        Continuous time foward dynamics evaluation
        
        dx = f(x,u,t)
        
        INPUTS
        x  : state vector             n x 1
        u  : control inputs vector    m x 1
        t  : time                     1 x 1
        
        OUPUTS
        dx : state derivative vectror n x 1
        
        """
        
        # from state vector (x) to angle and speeds (q,dq)
        [ q , dq ] = self.x2q( x )       
        
        # compute joint acceleration 
        ddq = self.ddq( q , dq , u , t ) 
        
        # from angle and speeds diff (dq,ddq) to state vector diff (dx)
        dx = self.q2x( dq , ddq )        
        
        return dx
    
    
    ###########################################################################
    def kinetic_energy(self, q  , dq ):
        """ Compute kinetic energy of manipulator """  
        
        e_k = 0.5 * np.dot( dq , np.dot( self.H( q ) , dq ) )
        
        return e_k



###############################################################################
###
###############################################################################


class MechanicalSystemWithPositionInputs( MechanicalSystem  ):
    """ 
    Mechanical system with Equation of Motion in the form of
    -------------------------------------------------------
    H(q) ddq + C(q,dq) dq + d(q,dq,u) + g(q) = B(q,u) e(u)
    -------------------------------------------------------
    q      :  dim = (dof, 1)   : position variables 
    dq     :  dim = (dof, 1)   : velocity variables     
    ddq    :  dim = (dof, 1)   : acceleration variables
    e(u)   :  dim = (m_f, 1)   : force input variables
    H(q)   :  dim = (dof, dof) : inertia matrix
    C(q)   :  dim = (dof, dof) : corriolis matrix
    B(q,u) :  dim = (dof, m_f) : actuator matrix
    ddq    :  dim = (dof, 1)   : acceleration variables
    d(q,dq):  dim = (dof, 1)   : state-dependent dissipative forces
    g(q)   :  dim = (dof, 1)   : state-dependent conservatives forces
    
    m = m_f + m_o
    ---------------------------------------------------
    m      :   integer         : number of inputs
    m_f    :   integer         : number of force inputs
    m_o    :   integer         : number of other inputs
    u      :   dim = ( m , 1)  : vector of all input variables
    
    """
    
    ############################
    def __init__(self, dof = 1 , force_inputs = 1, other_inputs = 1):
        """ """
        
        # Degree of Freedom
        self.dof = dof
        
        # Nb of actuators
        self.m_f = force_inputs
        self.m_o = other_inputs
        
        # Dimensions
        n = dof * 2 
        m = self.m_f + self.m_o
        p = dof * 2
        
        # initialize standard params
        system.ContinuousDynamicSystem.__init__(self, n, m, p)
        
        # Name
        self.name = str(dof) + 'DoF Mechanical System'
        
        # Labels, bounds and units
        for i in range(dof):
            # joint angle states
            self.x_ub[i] = + np.pi * 2
            self.x_lb[i] = - np.pi * 2
            self.state_label[i] = 'Angle '+ str(i)
            self.state_units[i] = '[rad]'
            # joint velocity states
            self.x_ub[i+dof] = + np.pi * 2
            self.x_lb[i+dof] = - np.pi * 2
            self.state_label[i+dof] = 'Velocity ' + str(i)
            self.state_units[i+dof] = '[rad/sec]'
        for i in range(self.m_f):
            self.u_ub[i] = + 5
            self.u_lb[i] = - 5
            self.input_label[i] = 'Force ' + str(i)
            self.input_units[i] ='[N]'
        self.output_label = self.state_label
        self.output_units = self.state_units
            
    ###########################################################################
    # The following functions needs to be overloaded by child classes
    # to represent the dynamic of the system
    ###########################################################################
    
    
    ###########################################################################
    def B(self, q , u ):
        """ 
        Actuator Matrix  : dof x m
        """
        
        B = np.zeros( ( self.dof , self.m_f ) )
        
        for i in range(min(self.m_f,self.dof)):
            B[i,i] = 1                # Diag matrix for the first m rows
        
        return B
    
    
    #############################
    def u2e( self, u ):
        """  """
        
        e = u[ 0 : self.m_f ] 
        
        return e
    
    
    ###########################################################################
    def d(self, q , dq , u ):
        """ 
        State-dependent dissipative forces : dof x 1
        """
        
        d = np.zeros(self.dof ) # Default is zero vector
        
        return d
    
    
    ###########################################################################
    # No need to overwrite the following functions for custom system
    ###########################################################################
    
    
    ##############################
    def generalized_forces(self, q  , dq  , ddq , t = 0 ):
        """ Computed generalized forces given a trajectory """  
        
        H = self.H( q )
        C = self.C( q , dq )
        g = self.g( q )
        
        u = self.ubar
        d = self.d( q , dq , u )
                
        # Generalized forces
        forces = np.dot( H , ddq ) + np.dot( C , dq ) + g + d
        
        return forces
    
    
    ##############################
    def actuator_forces(self, q  , dq  , ddq , t = 0 ):
        """ Computed actuator forces given a trajectory (inverse dynamic) """  
        
        raise NotImplementedError
        
    
    ##############################
    def ddq(self, q , dq , u , t = 0 ):
        """ Computed accelerations given actuator forces (foward dynamic) """  
        
        H = self.H( q )
        C = self.C( q , dq )
        g = self.g( q  )
        d = self.d( q , dq, u )
        
        B = self.B( q , u )
        e = self.u2e( u )
        
        ddq = np.dot( np.linalg.inv( H ) ,  ( np.dot( B , e )  
                                            - np.dot( C , dq ) - g - d ) )
        
        return ddq


    
'''
#################################################################
##################          Main                         ########
#################################################################
'''


if __name__ == "__main__":     
    """ MAIN TEST """
    
    sys = MechanicalSystem( 2 )
    
    sys.show(  q = np.array([ 1.0, 2.0]) )
    sys.show3( q = np.array([-0.5, 1.5]) )
    
    sys.ubar = np.array([1,2])
    sys.x0   = np.array([0,0,0,0])
    
    sys.plot_trajectory()
    sys.animate_simulation()
        


# -*- coding: utf-8 -*-
"""
Created on Tue Nov 13 11:05:07 2018

@author: Alexandre
"""

import numpy as np

from pyro.dynamic  import vehicle_steering
from pyro.planning import randomtree

sys  = vehicle_steering.HolonomicMobileRobotwithObstacles()

sys.obstacles = [
                [ (2,2),(4,10)],
                [ (6,-8),(8,8)],
                [ (-8,-8),(-1,8)]
                ]
        
x_start = np.array([8.0,0])
x_goal  = np.array([0.0,0])

planner = randomtree.RRT( sys , x_start )

v = 1

planner.u_options = [
        np.array([+v,+v]),
        np.array([+v, 0]),
        np.array([+v,-v]),
        np.array([-v,-v]),
        np.array([-v, 0]),
        np.array([-v,+v]),
        np.array([ 0,+v]),
        np.array([ 0,-v]),
        ]

planner.dt          = 0.1
planner.steps       = 5
planner.alpha       = 0.9
planner.beta        = 0.1
planner.goal_radius = 1.0

planner.find_path_to_goal( x_goal )

planner.plot_tree()
planner.show_solution()
planner.animate_solution()


# -*- coding: utf-8 -*-
"""
Created on Fri Nov 16 12:05:08 2018

@author: Alexandre
"""
###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic import pendulum
from pyro.control import nonlinear
###############################################################################

sys  = pendulum.SinglePendulum()


ctl  = nonlinear.SlidingModeController( sys )

ctl.lam  = 3.0
ctl.gain = 20.0

# Set Point
q_target = np.array([3.14])
ctl.rbar = q_target

# New cl-dynamic
cl_sys = ctl + sys

# Simultation
cl_sys.x0 = np.array([0.1,0])
cl_sys.compute_trajectory(tf=5, n=10001, solver='euler') 
# Note: Use "euler" solver when using sliding mode controllers
cl_sys.plot_trajectory('xu')
cl_sys.plot_phase_plane_trajectory_closed_loop()
cl_sys.plot_phase_plane()
cl_sys.animate_simulation()


# -*- coding: utf-8 -*-
"""
Created on Thu Nov  8 21:05:55 2018

@author: Alexandre
"""

import sys as python_system
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.animation as animation


from pyro.analysis import phaseanalysis


###############################################################################
#  Note: Modify here matplolib setting to fit your environment
###############################################################################

# Use interactive backend
try:
    # Default usage for interactive mode
    matplotlib.use('Qt5Agg')
    plt.ion() # Set interactive mode
    
except:
    pass

# Default figure settings
default_figsize   = (4, 3)
default_dpi       = 250
default_linestyle = '-'
default_fontsize  = 5

# True if running in IPython, False if running the file in terminal
if hasattr(python_system, 'ps1'): 
    figure_blocking  = False   # Set to not block the code when showing a figure
else:
    # We want to block figure to prevent the script from terminating
    figure_blocking  = True   # Set to block the code when showing a figure
        
# Embed font type in PDF when exporting
matplotlib.rcParams['pdf.fonttype'] = 42
matplotlib.rcParams['ps.fonttype']  = 42
    
    
    

###############################################################################
class TrajectoryPlotter:
    
    ##########################################################################
    def __init__(self, sys):
        
        self.sys = sys

        # Ploting
        self.fontsize = default_fontsize
        self.figsize  = default_figsize
        self.dpi      = default_dpi 


    ##########################################################################
    def plot(self, traj, plot = 'x' , show = True ):
        """
        Create a figure with trajectories for states, inputs, outputs and cost
        ----------------------------------------------------------------------
        plot = 'All'
        plot = 'xu'
        plot = 'xy'
        plot = 'x'
        plot = 'u'
        plot = 'y'
        plot = 'j'
        plot = 'z'
        """

        if 'j' in plot and (traj.J is None or traj.dJ is None):
            raise ValueError(
                "Trajectory does not contain cost data but plotting 'j' was requested"
            )

        sys = self.sys

        # For closed-loop systems, extract the inner Dynamic system for plotting
        #try:
        #    sys = self.sys.plant # sys is the global system
        #except AttributeError:
        #    pass

        # Number of subplots
        if plot == 'All':
            l = sys.n + sys.m + sys.p + 2
        elif plot == 'xuj':
            l = sys.n + sys.m + 2
        elif plot == 'xu':
            l = sys.n + sys.m
        elif plot == 'xy':
            l = sys.n + sys.p
        elif plot == 'x':
            l = sys.n
        elif plot == 'u':
            l = sys.m
        elif plot == 'y':
            l = sys.p
        elif plot == 'j':
            l = 2
        elif plot == 'z':
            l = sys.controller.l
        else:
            raise ValueError('not a valid ploting argument')

        simfig , plots = plt.subplots(l, sharex=True, figsize=self.figsize,
                                      dpi=self.dpi, frameon=True)
        
        lines = [None] * l

        #######################################################################
        #Fix bug for single variable plotting
        if l == 1:
            plots = [plots]
        #######################################################################

        simfig.canvas.manager.set_window_title('Trajectory for ' + self.sys.name)

        j = 0 # plot index

        if plot=='All' or plot=='x' or plot=='xu' or plot=='xy' or plot=='xuj':
            # For all states
            for i in range( sys.n ):
                lines[j] = plots[j].plot( traj.t , traj.x[:,i] , 'b')[0]
                plots[j].set_ylabel(sys.state_label[i] +'\n'+
                sys.state_units[i] , fontsize=self.fontsize )
                plots[j].grid(True)
                plots[j].tick_params( labelsize = self.fontsize )
                j = j + 1

        if plot == 'All' or plot == 'u' or plot == 'xu' or plot == 'xuj':
            # For all inputs
            for i in range( sys.m ):
                lines[j] = plots[j].plot( traj.t , traj.u[:,i] , 'r')[0]
                plots[j].set_ylabel(sys.input_label[i] + '\n' +
                sys.input_units[i] , fontsize=self.fontsize )
                plots[j].grid(True)
                plots[j].tick_params( labelsize = self.fontsize )
                j = j + 1

        if plot == 'All' or plot == 'y' or plot == 'xy':
            # For all outputs
            for i in range( sys.p ):
                lines[j] = plots[j].plot( traj.t , traj.y[:,i] , 'k')[0]
                plots[j].set_ylabel(sys.output_label[i] + '\n' +
                sys.output_units[i] , fontsize=self.fontsize )
                plots[j].grid(True)
                plots[j].tick_params( labelsize = self.fontsize )
                j = j + 1

        if plot == 'All' or plot == 'j' or plot == 'xuj':
            # Cost function
            lines[j] = plots[j].plot( traj.t , traj.dJ[:] , 'b')[0]
            plots[j].set_ylabel('dJ', fontsize=self.fontsize )
            plots[j].grid(True)
            plots[j].tick_params( labelsize = self.fontsize )
            j = j + 1
            lines[j] = plots[j].plot( traj.t , traj.J[:] , 'r')[0]
            plots[j].set_ylabel('J', fontsize=self.fontsize )
            plots[j].grid(True)
            plots[j].tick_params( labelsize = self.fontsize )
            j = j + 1
            
        if plot == 'z':
            # Internal states
            n = sys.n - sys.controller.l
            for i in range( l ):
                lines[j] = plots[j].plot( traj.t , traj.x[:,i+n] , 'b')[0]
                plots[j].set_ylabel(sys.state_label[i+n] +'\n'+
                sys.state_units[i+n] , fontsize=self.fontsize )
                plots[j].grid(True)
                plots[j].tick_params( labelsize = self.fontsize )
                j = j + 1
                
        plots[l-1].set_xlabel('Time [sec]', fontsize=self.fontsize )

        simfig.tight_layout()

        simfig.canvas.draw()
        plt.draw()
        plt.show( block = figure_blocking )

        self.fig   = simfig
        self.plots = plots
        self.lines = lines
        self.l     = l
        
        return (simfig, plots, lines)
    
    ##########################################################################
    def update_plot(self, traj, plot = 'x'):
        """
        Create a figure with trajectories for states, inputs, outputs and cost
        ----------------------------------------------------------------------
        plot = 'All'
        plot = 'xu'
        plot = 'xy'
        plot = 'x'
        plot = 'u'
        plot = 'y'
        plot = 'j'
        plot = 'z'
        """
        
        lines = self.lines
        sys   = self.sys
        plots = self.plots
        
        j = 0
        
        if plot=='All' or plot=='x' or plot=='xu' or plot=='xy' or plot=='xuj':
            
            # For all states
            for i in range( sys.n ):
                lines[j].set_data( traj.t , traj.x[:,i] )
                plots[j].relim()
                plots[j].autoscale_view(True,True,True)
                j = j + 1

        if plot == 'All' or plot == 'u' or plot == 'xu' or plot == 'xuj':
            # For all inputs
            for i in range( sys.m ):
                lines[j].set_data( traj.t , traj.u[:,i] )
                plots[j].relim()
                plots[j].autoscale_view(True,True,True)
                j = j + 1

        if plot == 'All' or plot == 'y' or plot == 'xy':
            # For all outputs
            for i in range( sys.p ):
                lines[j].set_data( traj.t , traj.y[:,i] )
                plots[j].relim()
                plots[j].autoscale_view(True,True,True)
                j = j + 1

        if plot == 'All' or plot == 'j' or plot == 'xuj':
            # Cost function
            lines[j].set_data( traj.t , traj.dJ[:,i] )
            plots[j].relim()
            plots[j].autoscale_view(True,True,True)
            j = j + 1
            lines[j].set_data( traj.t , traj.J[:,i] )
            plots[j].relim()
            plots[j].autoscale_view(True,True,True)
            j = j + 1
            
        if plot == 'z':
            # Internal states
            n = sys.n - sys.controller.l
            for i in range( self.l ):
                lines[j].set_data( traj.t , traj.x[:,i+n] )
                plots[j].relim()
                plots[j].autoscale_view(True,True,True)
                j = j + 1
        
    
    
    ##########################################################################
    def phase_plane_trajectory(self, traj, x_axis=0, y_axis=1):
        """ """
        pp = phaseanalysis.PhasePlot( self.sys , x_axis , y_axis )
        pp.plot()

        plt.plot(traj.x[:,x_axis], traj.x[:,y_axis], 'b-') # path
        plt.plot([traj.x[0,x_axis]], [traj.x[0,y_axis]], 'ko') # start
        plt.plot([traj.x[-1,x_axis]], [traj.x[-1,y_axis]], 'rx') # end
        
        plt.draw()

        pp.phasefig.tight_layout()
        
        plt.draw()
        plt.show( block = figure_blocking )
        

    ###########################################################################
    def phase_plane_trajectory_3d(self, traj, x_axis=0, y_axis=1, z_axis=2):
        """ """
        pp = phaseanalysis.PhasePlot3( self.sys , x_axis, y_axis, z_axis)

        pp.plot()

        pp.ax.plot(traj.x[:,x_axis],
                        traj.x[:,y_axis],
                        traj.x[:,z_axis],
                        'b-') # path
        pp.ax.plot([traj.x[0,x_axis]],
                        [traj.x[0,y_axis]],
                        [traj.x[0,z_axis]],
                        'o') # start
        pp.ax.plot([traj.x[-1,x_axis]],
                        [traj.x[-1,y_axis]],
                        [traj.x[-1,z_axis]],
                        's') # start # end

        pp.ax.set_xlim( self.sys.x_lb[ x_axis ] ,
                             self.sys.x_ub[ x_axis ])
        pp.ax.set_ylim( self.sys.x_lb[ y_axis ] ,
                             self.sys.x_ub[ y_axis ])
        pp.ax.set_zlim( self.sys.x_lb[ z_axis ] ,
                             self.sys.x_ub[ z_axis ])
        
        plt.draw()

        pp.phasefig.tight_layout()
        
        plt.draw()
        plt.show( block = figure_blocking )


    ###########################################################################
    def phase_plane_trajectory_closed_loop(self, traj, x_axis, y_axis):
        """ """
        pp = phaseanalysis.PhasePlot( self.sys , x_axis , y_axis )

        pp.compute_grid()
        pp.plot_init()

        # Closed-loop Behavior
        pp.color = 'r'
        pp.compute_vector_field()
        pp.plot_vector_field()

        # Open-Loop Behavior
        pp.f     = self.sys.f
        pp.ubar  = self.sys.ubar
        pp.color = 'b'
        pp.compute_vector_field()
        pp.plot_vector_field()

        pp.plot_finish()

        # Plot trajectory
        plt.plot(traj.x[:,x_axis], traj.x[:,y_axis], 'b-') # path
        plt.plot([traj.x[0,x_axis]], [traj.x[0,y_axis]], 'o') # start
        plt.plot([traj.x[-1,x_axis]], [traj.x[-1,y_axis]], 's') # end
        
        plt.draw()

        plt.tight_layout()
        
        plt.draw()
        plt.show( block = figure_blocking )
        
        
        
##########################################################################
# Animator
##########################################################################
class Animator:
    """ 

    """
    
    ###########################################################################
    def __init__(self, sys ):
        """
        
        sys = system.ContinuousDynamicSystem()
        
        sys needs to implement:
        
        get configuration from states, inputs and time
        ----------------------------------------------
        q             = sys.xut2q( x , u , t )
        
        get graphic output list of lines from configuration
        ----------------------------------------------
        lines_pts     = sys.forward_kinematic_lines( q )
        
        get graphic domain from configuration
        ----------------------------------------------
        ((,),(,),(,)) = sys.forward_kinematic_domain( q )
        
        """
        
        self.sys = sys
        
        self.x_axis = 0
        self.y_axis = 1
        
        # Ploting Param
        self.fontsize = default_fontsize
        self.figsize  = default_figsize
        self.dpi      = default_dpi 
        self.linestyle = sys.linestyle 
        
        # Label
        self.top_right_label = None
        

    ###########################################################################
    def show(self, q , x_axis = 0 , y_axis = 1 ):
        """ Plot figure of configuration q """
        
        # Update axis to plot in 2D
        
        self.x_axis = x_axis
        self.y_axis = y_axis
        
        # Get data
        lines      = self.sys.forward_kinematic_lines( q )
        domain     = self.sys.forward_kinematic_domain( q )
        
        # Two type of output are supported for foward_kinematic_lines
        # check if return is a tuple of ( pts , style , color )
        # or only the pts (then use default values)
        
        if type(lines) is tuple:
            # If the foward_kinematic_lines function specify style and color
            lines_pts   = lines[0]
            lines_style = lines[1]
            lines_color = lines[2]
        else:
            lines_pts   = lines
            lines_style = []
            lines_color = []
            for j, line in enumerate(lines):
                lines_style.append( self.sys.linestyle  )  # default value 
                lines_color.append( self.sys.linescolor )  # default value 
        
        # Plot
        self.showfig = plt.figure(figsize=self.figsize, dpi=self.dpi)
        self.showfig.canvas.manager.set_window_title('2D Configuration of ' + 
                                            self.sys.name )
        self.showax = self.showfig.add_subplot(111, autoscale_on=False )
        self.showax.grid()
        self.showax.axis('equal')
        self.showax.set_xlim(  domain[x_axis] )
        self.showax.set_ylim(  domain[y_axis] )
        self.showax.tick_params(axis='both', which='both', labelsize=self.fontsize)
        
        self.showlines = []
        
        for j , pts in enumerate(lines_pts):
            
            x_pts     = pts[:, x_axis ]
            y_pts     = pts[:, y_axis ]
            linestyle = lines_style[j] + lines_color[j]
            line,     = self.showax.plot( x_pts, y_pts, linestyle )
            
            self.showlines.append( line )

        plt.show( block = figure_blocking )
    
    
    ###########################################################################
    def show3(self, q ):
        """ Plot figure of configuration q """
        
        # Get data
        lines          = self.sys.forward_kinematic_lines( q )
        domain         = self.sys.forward_kinematic_domain( q )
        
        if type(lines) is tuple:
            # If the foward_kinematic_lines function specify style and color
            lines_pts   = lines[0]
            lines_style = lines[1]
            lines_color = lines[2]
        else:
            lines_pts   = lines
            lines_style = []
            lines_color = []
            for j, line in enumerate(lines):
                lines_style.append( self.sys.linestyle  )  # default value 
                lines_color.append( self.sys.linescolor )  # default value 
        
        # Plot
        self.show3fig = plt.figure(figsize=self.figsize, dpi=self.dpi)
        self.show3fig.canvas.manager.set_window_title('3D Configuration of ' + 
                                            self.sys.name )
        self.show3ax = self.show3fig.add_subplot(projection='3d')
                
        self.show3lines = []
        
        for j , pts in enumerate(lines_pts):
            x_pts     = pts[:, 0 ]
            y_pts     = pts[:, 1 ]
            z_pts     = pts[:, 2 ]
            linestyle = lines_style[j] + lines_color[j]
            line,     = self.show3ax.plot( x_pts, y_pts, z_pts, linestyle)
            
            self.show3lines.append( line )
            
        self.show3ax.set_xlim3d( domain[0] )
        self.show3ax.set_xlabel('X')
        self.show3ax.set_ylim3d( domain[1] )
        self.show3ax.set_ylabel('Y')
        self.show3ax.set_zlim3d( domain[2] )
        self.show3ax.set_zlabel('Z')
        self.show3ax.tick_params(axis='both', which='both', labelsize=self.fontsize)
        
        plt.show( block = figure_blocking )
        
    ###########################################################################
    def show_plus(self, x , u , t , x_axis = 0 , y_axis = 1 ):
        """ Plot figure of system at state x """
        
        # Update axis to plot in 2D
        
        self.x_axis = x_axis
        self.y_axis = y_axis
        
        # Get data
        lines_data = self.get_lines(x, u, t)
            
        # Save data in lists for the whole trajectory
        lines_pts        = lines_data[0]  
        lines_style      = lines_data[1]  
        lines_color      = lines_data[2]
        lines_plus_pts   = lines_data[3]
        lines_plus_style = lines_data[4]
        lines_plus_color = lines_data[5]
        domain           = lines_data[6]  
        
        # Plot
        self.showfig = plt.figure(figsize=self.figsize, dpi=self.dpi)
        self.showfig.canvas.manager.set_window_title('2D plot of ' + 
                                            self.sys.name )
        self.showax = self.showfig.add_subplot(111, autoscale_on=False )
        self.showax.grid()
        self.showax.axis('equal')
        self.showax.set_xlim(  domain[x_axis] )
        self.showax.set_ylim(  domain[y_axis] )
        self.showax.tick_params(axis='both', which='both', labelsize=self.fontsize)
        
        self.showlines      = []
        self.showlines_plus = []
        
        # for each lines
        for j, line_pts in enumerate( lines_pts ):
            
            linestyle = lines_style[j] + lines_color[j]
            
            thisx = line_pts[:,self.x_axis]
            thisy = line_pts[:,self.y_axis]
            line, = self.showax.plot(thisx, thisy, linestyle)

            self.showfig.tight_layout()
                
            self.showlines.append( line )
            
        # Lines plus optionnal 
        if self.sys.lines_plus:
            
            for j, line_pts in enumerate( lines_plus_pts ):
                
                linestyle = lines_plus_style[j] + lines_plus_color[j]
                
                thisx = line_pts[:,self.x_axis]
                thisy = line_pts[:,self.y_axis]
                line, = self.showax.plot(thisx, thisy, linestyle)
                    
                self.showlines_plus.append( line )

        plt.show( block = figure_blocking )
        
    ###########################################################################
    def show_plus_update(self, x , u , t ):
        """ Update a show plus figure """
        
        # Get data
        lines_data = self.get_lines(x, u, t)
            
        # Line data
        lines_pts        = lines_data[0]  
        lines_style      = lines_data[1]  
        lines_color      = lines_data[2]
        lines_plus_pts   = lines_data[3]
        lines_plus_style = lines_data[4]
        lines_plus_color = lines_data[5]
        domain           = lines_data[6]  
        
        # Update lines
        for j, line in enumerate(self.showlines):
            thisx = lines_pts[j][:,self.x_axis]
            thisy = lines_pts[j][:,self.y_axis]
            line.set_data(thisx, thisy)
            
        if self.sys.lines_plus:
            # Update lines plus
            for j, line in enumerate(self.showlines_plus):
                thisx = lines_plus_pts[j][:,self.x_axis]
                thisy = lines_plus_pts[j][:,self.y_axis]
                line.set_data(thisx, thisy)
            
        # Update domain
        i_x = self.x_axis
        i_y = self.y_axis
        self.showax.set_xlim( domain[i_x] )
        self.showax.set_ylim( domain[i_y] )
        
        self.showfig.canvas.draw()

        plt.show( block = figure_blocking )
        
        
    ###########################################################################
    def get_lines(self, x , u , t ):
        """ 
        shorcut to get all graphic output data
        """
        
        # Get configuration q from simulation
        q               = self.sys.xut2q( x , u , t )
        
        # Compute graphical forward kinematic
        lines       = self.sys.forward_kinematic_lines( q )
        lines_plus  = self.sys.forward_kinematic_lines_plus( x , u , t )
        domain      = self.sys.forward_kinematic_domain( q )
        
        
        # Two type of output are supported for foward_kinematic_lines
        # check if return is a tuple of ( pts , style , color )
        # or only the pts (then use default values)
        
        if type(lines) is tuple:
            # If the foward_kinematic_lines function specify style and color
            lines_pts   = lines[0]
            lines_style = lines[1]
            lines_color = lines[2]
        else:
            lines_pts   = lines
            lines_style = []
            lines_color = []
            for j, line in enumerate(lines):
                lines_style.append( self.sys.linestyle  )  # default value 
                lines_color.append( self.sys.linescolor )  # default value 
        
        if type(lines_plus) is tuple:
            # If the foward_kinematic_lines function specify style and color
            lines_plus_pts   = lines_plus[0]
            lines_plus_style = lines_plus[1]
            lines_plus_color = lines_plus[2]
        else:
            lines_plus_pts   = lines_plus
            lines_plus_style = []
            lines_plus_color = []
            for j, line_plus in enumerate(lines_plus):
                lines_plus_style.append( self.sys.linestyle_plus  )  # default value 
                lines_plus_color.append( self.sys.linescolor_plus )  # default value 
            
        lines_data = ( lines_pts ,lines_style , lines_color , lines_plus_pts ,
                       lines_plus_style , lines_plus_color , domain )
            
        return lines_data

        

    ###########################################################################
    def animate_simulation(self, traj, time_factor_video =  1.0 , is_3d = False, 
                                 save = False , file_name = 'Animation' , 
                                 show = True ):
        """ 
        Show Animation of the simulation 
        ----------------------------------
        time_factor_video < 1 --> Slow motion video        
        
        """  
        self.is_3d = is_3d
        
        # Init lists
        self.ani_lines_pts        = []
        self.ani_lines_style      = []
        self.ani_lines_color      = []
        
        self.ani_lines_plus_pts   = []
        self.ani_lines_plus_style = []
        self.ani_lines_plus_color = []
        
        self.ani_domains          = []

        nsteps = traj.t.size
        self.sim_dt = (traj.t[-1] - traj.t[0]) / (traj.t.size - 1)

        # For all simulation data points
        for i in range( nsteps ):
            
            # Get states , actions , time
            x = traj.x[i,:]
            u = traj.u[i,:]
            t = traj.t[i]
            
            lines_data = self.get_lines(x, u, t)
            
            # Save data in lists for the whole trajectory
            self.ani_lines_pts.append(        lines_data[0]  )
            self.ani_lines_style.append(      lines_data[1]  )
            self.ani_lines_color.append(      lines_data[2]  )
            self.ani_lines_plus_pts.append(   lines_data[3]  )
            self.ani_lines_plus_style.append( lines_data[4]  )
            self.ani_lines_plus_color.append( lines_data[5]  )
            self.ani_domains.append(          lines_data[6]  )
            
        # Init figure
        self.ani_fig = plt.figure(figsize=self.figsize, dpi=self.dpi )
        
        
        if is_3d:
            #self.ani_ax = p3.Axes3D(self.ani_fig) #TODO
            #self.ani_fig.add_axes(self.ani_ax)
            self.ani_ax = self.ani_fig.add_subplot(projection='3d')
            self.ani_ax.set_xlim3d(self.ani_domains[0][0])
            self.ani_ax.set_xlabel('X')
            self.ani_ax.set_ylim3d(self.ani_domains[0][1])
            self.ani_ax.set_ylabel('Y')
            self.ani_ax.set_zlim3d(self.ani_domains[0][2])
            self.ani_ax.set_zlabel('Z')
            self.ani_fig.canvas.manager.set_window_title('3D Animation of ' + 
                                            self.sys.name )
        else:
            self.ani_ax = self.ani_fig.add_subplot(111, autoscale_on=True)
            self.ani_ax.axis('equal')
            self.ani_ax.set_xlim(  self.ani_domains[0][self.x_axis] )
            self.ani_ax.set_ylim(  self.ani_domains[0][self.y_axis] )
            self.ani_fig.canvas.manager.set_window_title('2D Animation of ' + 
                                            self.sys.name )
            
        self.ani_ax.tick_params(axis='both', which='both', labelsize=
                                self.fontsize)
        self.ani_ax.grid()
                
        # Plot lines at t=0
        self.lines      = []
        self.lines_plus = []
        
        # for each lines of the t=0 data point
        for j, line_pts in enumerate( self.ani_lines_pts[0] ):
            
            linestyle = self.ani_lines_style[0][j] + self.ani_lines_color[0][j]
            
            if is_3d:
                thisx     = line_pts[:,0]
                thisy     = line_pts[:,1]
                thisz     = line_pts[:,2]
                
                line, = self.ani_ax.plot(thisx, thisy, thisz, linestyle)
                self.time_text = self.ani_ax.text(0, 0, 0, 'time =', 
                                                  transform=
                                                  self.ani_ax.transAxes)
                self.label_text = self.ani_ax.text(0.9, 0.9, 0.9, self.top_right_label)
                
            else:
                thisx = line_pts[:,self.x_axis]
                thisy = line_pts[:,self.y_axis]
                line, = self.ani_ax.plot(thisx, thisy, linestyle)
                self.time_text = self.ani_ax.text(0.05, 0.9, 'time =', 
                                                  transform=self.
                                                  ani_ax.transAxes)
                self.label_text = self.ani_ax.text(0.75, 0.8, self.top_right_label, 
                                                   transform=self.
                                                   ani_ax.transAxes)
                self.ani_fig.tight_layout()
                
            self.lines.append( line )
            
        # Lines plus optionnal 
        if self.sys.lines_plus:
            
            for j, line_pts in enumerate( self.ani_lines_plus_pts[0] ):
                
                linestyle = self.ani_lines_plus_style[0][j] + self.ani_lines_plus_color[0][j]
                
                if is_3d:
                    thisx     = line_pts[:,0]
                    thisy     = line_pts[:,1]
                    thisz     = line_pts[:,2]
                    line,     = self.ani_ax.plot(thisx, thisy, thisz, linestyle)
                    
                else:
                    thisx = line_pts[:,self.x_axis]
                    thisy = line_pts[:,self.y_axis]
                    line, = self.ani_ax.plot(thisx, thisy, linestyle)
                    
                self.lines_plus.append( line )
        
        self.time_template = 'time = %.1fs'
        
        
        # Animation
        inter      =  40.             # ms --> 25 frame per second
        frame_dt   =  inter / 1000. 
        
        if ( frame_dt * time_factor_video )  < self.sim_dt :
            # Simulation is slower than video
            
            # don't skip steps
            self.skip_steps = 1
            
            # adjust frame speed to simulation                                    
            inter           = self.sim_dt * 1000. / time_factor_video 
            
            n_frame         = nsteps
            
        else:
            # Simulation is faster than video
            
            # --> number of simulation frame to skip between video frames
            factor          =  frame_dt / self.sim_dt * time_factor_video
            self.skip_steps =  int( factor  ) 
            
            # --> number of video frames
            n_frame         =  int( nsteps / self.skip_steps )
        
        # ANIMATION
        # blit=True option crash on mac
        #self.ani = animation.FuncAnimation( self.ani_fig, self.__animate__, 
        # n_frame , interval = inter , init_func=self.__ani_init__ , blit=True)
        
        if self.is_3d:
            self.ani = animation.FuncAnimation( self.ani_fig, self.__animate__, 
                                                n_frame , interval = inter )
        else:
            self.ani = animation.FuncAnimation( self.ani_fig, self.__animate__,
                                                n_frame , interval = inter, 
                                                init_func=self.__ani_init__ )
        if save:
            self.ani.save( file_name + '.gif', writer='imagemagick', fps=30)

        # self.ani_fig.show( block = figure_blocking )
        if show:
            #plt.ioff()
            plt.show( block = figure_blocking )
            
        else:
            plt.close(self.ani_fig)
            
        return self.ani
        

    #####################################    
    def __ani_init__(self):
        
        for line in self.lines:
            line.set_data([], [])
            
        for line in self.lines_plus:
            line.set_data([], [])
            
        self.time_text.set_text('')
        
        return self.lines, self.lines_plus, self.time_text, self.ani_ax
    
    
    ######################################
    def __animate__(self,i):
        
        # Update lines
        for j, line in enumerate(self.lines):
            if self.is_3d:
                thisx = self.ani_lines_pts[i * self.skip_steps][j][:,0]
                thisy = self.ani_lines_pts[i * self.skip_steps][j][:,1]
                thisz = self.ani_lines_pts[i * self.skip_steps][j][:,2]
                line.set_data(thisx, thisy)
                line.set_3d_properties(thisz)
            else:
                thisx = self.ani_lines_pts[i*self.skip_steps][j][:,self.x_axis]
                thisy = self.ani_lines_pts[i*self.skip_steps][j][:,self.y_axis]
                line.set_data(thisx, thisy)
        
        if self.sys.lines_plus:
            # Update lines plus
            for j, line in enumerate(self.lines_plus):
                if self.is_3d:
                    thisx = self.ani_lines_plus_pts[i * self.skip_steps][j][:,0]
                    thisy = self.ani_lines_plus_pts[i * self.skip_steps][j][:,1]
                    thisz = self.ani_lines_plus_pts[i * self.skip_steps][j][:,2]
                    line.set_data(thisx, thisy)
                    line.set_3d_properties(thisz)
                else:
                    thisx = self.ani_lines_plus_pts[i*self.skip_steps][j][:,self.x_axis]
                    thisy = self.ani_lines_plus_pts[i*self.skip_steps][j][:,self.y_axis]
                    line.set_data(thisx, thisy)
            
        # Update time
        self.time_text.set_text(self.time_template % 
                                ( i * self.skip_steps * self.sim_dt )
                                )
        
        # Update domain
        if self.is_3d:
            self.ani_ax.set_xlim3d( self.ani_domains[i * self.skip_steps][0] )
            self.ani_ax.set_ylim3d( self.ani_domains[i * self.skip_steps][1] )
            self.ani_ax.set_zlim3d( self.ani_domains[i * self.skip_steps][2] )
        else:
            i_x = self.x_axis
            i_y = self.y_axis
            self.ani_ax.set_xlim( self.ani_domains[i * self.skip_steps][i_x] )
            self.ani_ax.set_ylim( self.ani_domains[i * self.skip_steps][i_y] )
        
        return self.lines, self.lines_plus, self.time_text, self.ani_ax
    

'''
###############################################################################
##################          Main                         ######################
###############################################################################
'''


if __name__ == "__main__":     
    """ MAIN TEST """
    
    from pyro.dynamic import pendulum
    from pyro.dynamic import vehicle_steering
    
    sys    = pendulum.DoublePendulum()
    sys.x0 = np.array([0.1,0.1,0,0])
    
    sys.plot_phase_plane(0,2)
    
    traj = sys.compute_trajectory( 2.0 )
    
    
    plotter = TrajectoryPlotter( sys )
    plotter.plot( traj, 'xu' )
    
    sys.x0 = np.array([0.2,0.1,0,0])
    traj2 = sys.compute_trajectory( 2.0  )
    
    plotter.update_plot( traj2 )
    
    plotter.phase_plane_trajectory( traj2 , 0 , 2 )

    is_3d = False
    
    sys.plot_trajectory()
    
    a = Animator(sys)
    a.animate_simulation( sys.traj, 1, is_3d)
    
    sys      = vehicle_steering.KinematicBicyleModel()
    sys.ubar = np.array([1,0.01])
    sys.x0   = np.array([0,0,0])
    
    b = Animator(sys)
    b.top_right_label = 'this is a label'
    sys.compute_trajectory( 100 )
    sys.plot_trajectory()
    b.animate_simulation( sys.traj, 10, is_3d)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Oct  4 05:49:06 2021

@author: alex
"""

import numpy as np
import matplotlib.pyplot as plt

import time

from scipy.optimize import minimize

from pyro.analysis import simulation
from pyro.planning import plan
from pyro.analysis import graphical


###############################################################################
class DirectCollocationTrajectoryOptimisation( plan.Planner ):
    """ 
    Trajectory optimisation based on fixed-time-steps direct collocation
    ---------------------------------------------------------------------
    sys  : dynamical system class
    dt   : time step size
    grid : number of time step (discretization number)
    cf   : cost function class
    
    
    """
    
    ############################
    def __init__(self, sys , dt = 0.2 , grid = 20 , cost_function = None,  dynamic_plot = False ):
        
        
        # Set sys, default cost function x_start and x_goal
        plan.Planner.__init__(self, sys , cost_function )
        
        # Discretization parameters
        self.dt    = dt
        self.grid  = grid
        
        # Parameters
        self.EPS     = 0.01
        self.maxiter = 100
        
        # Initial computation
        self.compute_bounds()
        self.dec_init   = np.zeros( grid * ( sys.n + sys.m ) )
        
        
        # Check if vectorized operation are available
        try:
            is_vectorized = self.sys.is_vectorized & self.cost_function.is_vectorized
        except:
            is_vectorized = False
        self.is_vectorized = is_vectorized
            
        
        
        # Memory variable
        self.iter_count = 0
        self.start_time = time.time()
        
        # Optional Convergence Graph
        self.dynamic_plot = dynamic_plot
        
        if dynamic_plot:
            
            self.init_dynamic_plot()
            
            
    ############################
    def init_dynamic_plot(self):
        
        # Graphic option
        self.dynamic_plot = True
            
        traj = self.decisionvariables2traj( self.dec_init )
        self.plotter = graphical.TrajectoryPlotter( self.sys )
        self.plotter.plot( traj, 'xu' )
        
        
    ############################
    def set_initial_trajectory_guest(self, traj):
        
        new_traj      = traj.re_sample( self.grid )
        self.dec_init = self.traj2decisionvariables( new_traj )
        
        if self.dynamic_plot:
            
            traj = self.decisionvariables2traj( self.dec_init )
            self.plotter.update_plot( traj, 'xu' )
            plt.pause( 0.001 )
            
    ############################
    def set_linear_initial_guest(self, derivatives = False ):
        
        xs = np.linspace(  self.x_start,   self.x_goal, self.grid )
        us = np.linspace( self.sys.ubar, self.sys.ubar, self.grid )
        
        # For second order mechanical system
        if derivatives:
            dof = int(self.sys.n/2)
            tf  = self.grid * self.dt
            dx  = ( self.x_goal[:dof] - self.x_start[:dof] ) / tf
            dxs = np.linspace( dx, dx, self.grid )
            xs[:,dof:] = dxs
        
        
        dec = np.array([]).reshape(0,1) # initialize dec_vars array
        
        for i in range(self.sys.n): # append states x
            arr_to_add = xs[:,i].reshape(self.grid,1)
            dec = np.append(dec,arr_to_add,axis=0)
    
        for i in range(self.sys.m): # append inputs u
            arr_to_add = us[:,i].reshape(self.grid,1)
            dec = np.append(dec,arr_to_add,axis=0)
            
        self.dec_init = dec[:,0]
        
        
        if self.dynamic_plot:
            
            traj = self.decisionvariables2traj( self.dec_init )
            self.plotter.update_plot( traj, 'xu' )
            plt.pause( 0.001 )
        
    
    ############################
    def decisionvariables2xu(self, dec ):
        """ 
        Unpack decision variable vector into x and u trajectory matrices 
        --------------------------
        
        dec = [ x[0](t=0), .... x[0](t), .... x[0](t=tf), 
               ...
               x[i](t=0), .... x[i](t), .... x[i](t=tf), 
               ...
               x[n](t=0), .... x[n](t), .... x[n](t=tf), 
               
               u[0](t=0), .... u[0](t), .... u[0](t=tf), 
               ...
               u[j](t=0), .... u[j](t), .... u[j](t=tf), 
               ...
               u[m](t=0), .... u[m](t), .... u[m](t=tf) ]
        
        """
        
        n    = self.sys.n   # number of states
        m    = self.sys.m   # number of inputs
        grid = self.grid    # number of time steps
    
        # x = np.zeros( ( n , grid ) ) 
        # u = np.zeros( ( m , grid ) )
        
        # # Extract states variables
        # for i in range(self.sys.n):
        #     x[i,:] = dec[ i * grid : (i+1) * grid ]
            
        # # Extract input variables
        # for j in range(self.sys.m):
        #     k = n + j
        #     u[j,:] = dec[ k * grid : (k+1) * grid ]
        
        x = dec[: n * grid ].reshape( n , grid )
        u = dec[ n * grid :].reshape( m , grid )
        
        return x,u
    
    
    ############################
    def traj2decisionvariables(self, traj):
        """ 
        Compute decision variables based onna trajectory object
        --------------------------
        
        dec = [ x[0](t=0), .... x[0](t), .... x[0](t=tf), 
               ...
               x[i](t=0), .... x[i](t), .... x[i](t=tf), 
               ...
               x[n](t=0), .... x[n](t), .... x[n](t=tf), 
               
               u[0](t=0), .... u[0](t), .... u[0](t=tf), 
               ...
               u[j](t=0), .... u[j](t), .... u[j](t=tf), 
               ...
               u[m](t=0), .... u[m](t), .... u[m](t=tf) ]
        
        """
        
        #n = grid*(self.sys.n+self.sys.m)
        
        dec = np.array([]).reshape(0,1) # initialize dec_vars array
        
        for i in range(self.sys.n): # append states x
            arr_to_add = traj.x[:,i].reshape(self.grid,1)
            dec = np.append(dec,arr_to_add,axis=0)
    
        for i in range(self.sys.m): # append inputs u
            arr_to_add = traj.u[:,i].reshape(self.grid,1)
            dec = np.append(dec,arr_to_add,axis=0)
        
        return dec[:,0]
    
    
    ############################
    def decisionvariables2traj(self, dec):
        """ 
        Unpack decision variable vector into x and u trajectory matrices 
        --------------------------
        
        dec = [ x[0](t=0), .... x[0](t), .... x[0](t=tf), 
               ...
               x[i](t=0), .... x[i](t), .... x[i](t=tf), 
               ...
               x[n](t=0), .... x[n](t), .... x[n](t=tf), 
               
               u[0](t=0), .... u[0](t), .... u[0](t=tf), 
               ...
               u[j](t=0), .... u[j](t), .... u[j](t=tf), 
               ...
               u[m](t=0), .... u[m](t), .... u[m](t=tf) ]
        
        """
        
        n    = self.sys.n   # number of states
        m    = self.sys.m   # number of inputs
        p    = self.sys.p   # number of inputs
        grid = self.grid    # number of time steps
    
        x = np.zeros((grid,n)) 
        u = np.zeros((grid,m))
        t  = np.zeros(grid)
        y  = np.zeros(( grid, p ))
        dx = np.zeros(( grid, n ))
        dJ = np.zeros(grid)
        J  = np.zeros(grid)
        
        J_sum = 0
        
        xT,uT = self.decisionvariables2xu( dec )
        
        for i in range(self.grid):
            x[i,:]  = xT[:,i]
            u[i,:]  = uT[:,i]
            t[i]    = i*self.dt
            dx[i,:] = self.sys.f(x[i,:],u[i,:],t[i])
            dJ[i]   = self.cost_function.g(x[i,:],u[i,:],t[i])
            
            J_sum = J_sum + dJ[i]
            J[i]  = J_sum
            
        #########################
        traj = simulation.Trajectory(
          x = x,
          u = u,
          t = t,
          dx= dx,
          y = y,
          dJ = dJ,
          J  = J
        )
        
        self.traj = traj
        
        return traj
        
    
    ############################
    def cost(self, dec):
        """ Compute cost for given decision variable using trapez integration approx """
        
        x,u = self.decisionvariables2xu( dec )
        
        if self.is_vectorized:
            
            # Vectorized operation version
            t  = np.linspace(0, ( self.grid - 1 )* self.dt, self.grid)
            
            dJ = self.cost_function.g( x , u , t )
            
            J  = np.trapz( dJ , t )
            
        else:
            
            # Loop version
        
            J = 0
            
            for i in range(self.grid -1):
                #i
                x_i = x[:,i]
                u_i = u[:,i]
                t_i = i*self.dt
                dJi = self.cost_function.g( x_i , u_i, t_i )
                
                #i+1
                x_i1 = x[:,i+1]
                u_i1 = u[:,i+1]
                t_i1 = (i+1)*self.dt
                dJi1 = self.cost_function.g( x_i1 , u_i1, t_i1 )
                
                #trapez
                dJ = 0.5 * ( dJi + dJi1 )
                
                #integral
                J = J + dJ * self.dt
            
        return J
    
    
    ########################
    def dynamic_constraints(self, dec):
        """ Compute residues of dynamic constraints """
    
        x , u = self.decisionvariables2xu( dec )
        
        if self.is_vectorized:
            
            # Vectorized operation version
            
            t  = np.linspace(0, ( self.grid - 1 )* self.dt, self.grid)
            
            dx = self.sys.f( x ,u , t )
            
            dx_eqs = 0.5 * ( dx[:,0:-1] + dx[:,1:] ) * self.dt
            
            dx_num = np.diff( x )
            
            residues = dx_num - dx_eqs
            
        
        else:
            
            # Loop version
        
            residues = np.zeros( ( self.grid - 1 , self.sys.n  ))
            
            for i in range(self.grid-1):
                
                #i
                x_i = x[:,i]
                u_i = u[:,i]
                t_i = i*self.dt
                dx_i = self.sys.f(x_i,u_i,t_i) # analytical state derivatives
                
                #i+1
                x_i1 = x[:,i+1]
                u_i1 = u[:,i+1]
                t_i1 = (i+1)*self.dt
                dx_i1 = self.sys.f(x_i1,u_i1,t_i1) # analytical state derivatives
                
                #trapez
                delta_x_eqs = 0.5 * self.dt * (dx_i + dx_i1)
                
                #num diff
                delta_x_num = x[:,i+1] - x[:,i] # numerical delta in trajectory data
                
                residues[i,:] = delta_x_num - delta_x_eqs
            
        return residues.flatten()
    
    
    ##############################
    def compute_bounds(self):
        """ Compute lower and upper bound vector for all decision variables """
    
        bounds = []
        
        # States constraints
        for j in range(self.sys.n):
            
            # initial states
            bounds.append( ( self.x_start[j] , self.x_start[j] + self.EPS ) )
            
            # range for intermediate states
            for i in range(1,self.grid - 1 ):
                bounds.append( ( self.sys.x_lb[j] , self.sys.x_ub[j] ) )
                
            # final goal state
            bounds.append( ( self.x_goal[j] , self.x_goal[j] + self.EPS ) )
        
        
        # Ipnut constraints
        for j in range(self.sys.m):

            for i in range(0,self.grid):
                bounds.append( ( self.sys.u_lb[j] , self.sys.u_ub[j] ) )
        
            
        self.bounds = bounds
    
    
    ##############################
    def display_callback(self, x ):
        
        self.iter_count = self.iter_count + 1
        
        print('Optimizing trajectory: iteration no', self.iter_count , 
              ' elapsed time = %.2f' % (time.time() - self.start_time ) )
        
        if self.dynamic_plot:
            
            traj = self.decisionvariables2traj( x )
            self.plotter.update_plot( traj, 'xu' )
            plt.pause( 0.001 )
        
        
        
    ##############################
    def compute_optimal_trajectory(self):
        
        self.start_time = time.time()
        
        self.compute_bounds()
        
        dynamic_constraints = {'type': 'eq', 'fun': self.dynamic_constraints }
    
        res = minimize(self.cost, 
                       self.dec_init, 
                       method='SLSQP',  
                       bounds=self.bounds, 
                       constraints=dynamic_constraints, 
                       callback=self.display_callback, 
                       options={'disp':True,'maxiter':self.maxiter})
        
        self.res  = res
        self.traj = self.decisionvariables2traj( self.res.x )
        
        
    ##############################
    def compute_solution(self):
        
        self.compute_optimal_trajectory()
        
        return self.traj
    


'''
#################################################################
##################          Main                         ########
#################################################################
'''


if __name__ == "__main__":     
    """ MAIN TEST """
    
    from pyro.dynamic import pendulum
    

    sys  = pendulum.SinglePendulum()
    
    planner = DirectCollocationTrajectoryOptimisation( sys )
    
    planner.x_start = np.array([0.1,0])
    planner.x_goal  = np.array([-3.14,0])
    
    planner.init_dynamic_plot()
    
    planner.compute_solution()
    planner.animate_solution()

    
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Oct  3 08:27:06 2021

@author: alex
"""

import numpy as np

from pyro.dynamic.drone                   import Drone2D
from pyro.planning.trajectoryoptimisation import DirectCollocationTrajectoryOptimisation


sys = Drone2D()

sys.u_ub = np.array([10.0, 10.0])

planner = DirectCollocationTrajectoryOptimisation( sys )

planner.x_start = np.array([-5,0,0,0,0,0])
planner.x_goal  = np.array([0,0,0,0,0,0])

planner.maxiter = 200
planner.init_dynamic_plot()
planner.compute_optimal_trajectory()
# planner.show_solution()
planner.animate_solution()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Oct 16 22:27:47 2022

@author: alex
"""

import numpy as np

from pyro.dynamic  import vehicle_propulsion
from pyro.analysis import costfunction
from pyro.planning import dynamicprogramming 
from pyro.planning import discretizer

sys  = vehicle_propulsion.LongitudinalFrontWheelDriveCarWithWheelSlipInput()

sys.x_ub[1] = 15
sys.x_lb[1] = 0

# Discrete world 
grid_sys = discretizer.GridDynamicSystem( sys , [101,101] , [11] , 0.05 )

# Cost Function
qcf = costfunction.QuadraticCostFunction.from_sys( sys )

qcf.xbar = np.array([ 45 , 0 ]) # target
qcf.Q[0,0] = 0.1
qcf.Q[1,1] = 0.1
qcf.INF  = 1000000

# DP algo
dp = dynamicprogramming.DynamicProgrammingWithLookUpTable( grid_sys, qcf)

dp.solve_bellman_equation( tol = 0.5 , animate_cost2go = True )


ctl = dp.get_lookup_table_controller()

ctl.plot_control_law( sys = sys , n = 100)


# Simulation and animation
cl_sys = ctl + sys
cl_sys.x0   = np.array([0,0])
cl_sys.compute_trajectory( 10, 10001, 'euler')
cl_sys.plot_trajectory('xu')
cl_sys.plot_phase_plane_trajectory()
cl_sys.animate_simulation()


# -*- coding: utf-8 -*-
"""
Created on Tue Oct 23 20:45:37 2018

@author: Alexandre
"""


###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic import system
from pyro.kinematic import geometry
from pyro.kinematic import drawing
###############################################################################

###############################################################################
class GeneralizedMechanicalSystem( system.ContinuousDynamicSystem ):
    """ 
    Mechanical system where the generalized velocities are not the time 
    derivative of the generalized coordinates.
    -------------------------------------------------------
    M(q) dv + C(q,v) v + d(q,v) + g(q) = B(q) u
    dq = N(q) v
    -------------------------------------------------------
    v        :  dim = (dof, 1)   : velocity variables
    q        :  dim = (pos, 1)   : position variables 
    dq       :  dim = (pos, 1)   : derivatives of position variables  
    dv       :  dim = (dof, 1)   : acceleration variables
    u        :  dim = (m, 1)     : force input variables
    d(q,v)   :  dim = (dof, 1)   : state-dependent dissipative forces
    g(q)     :  dim = (dof, 1)   : state-dependent conservatives forces
    M(q)     :  dim = (dof, dof) : inertia matrix
    C(q,v)   :  dim = (dof, dof) : corriolis matrix
    B(q)     :  dim = (dof, m)   : actuator matrix
    N(q)     :  dim = (pos, dof) : transformation matrix
    
    """
    
    ############################
    def __init__(self, dof = 1 , pos = None, actuators = None):
        """ """
        
        # Degree of Freedom
        self.dof = dof

        # Nb of configurations
        if pos == None:   # If not specifyied 
            pos = dof
        self.pos = pos
        
        # Nb of actuators
        if actuators == None:   # If not specifyied the sys is fully actuated
            actuators = dof
        
        # Dimensions
        n = dof + pos
        m = actuators
        p = n
        
        # initialize standard params
        system.ContinuousDynamicSystem.__init__(self, n, m, p)
        
        # Name
        self.name = str(dof) + 'DoF Generalized Mechanical System'
        
        # Labels, bounds and units
        for i in range(pos):
            # positions states
            self.x_ub[i] = + 10
            self.x_lb[i] = - 10
            self.state_label[i] = 'Position '+ str(i)
            self.state_units[i] = '[m]'
        for i in range(dof):
            # generalized velocity states
            j = i + pos
            self.x_ub[j] = + 10
            self.x_lb[j] = - 10
            self.state_label[j] = 'Velocity ' + str(i)
            self.state_units[j] = '[m/sec]'
        for i in range(actuators):
            self.u_ub[i] = + 5
            self.u_lb[i] = - 5
            self.input_label[i] = 'Force input ' + str(i)
            self.input_units[i] ='[N]'
        self.output_label = self.state_label
        self.output_units = self.state_units
            
    ###########################################################################
    # The following functions needs to be overloaded by child classes
    # to represent the dynamic of the system
    ###########################################################################
    
    ###########################################################################
    def M(self, q ):
        """ 
        Inertia matrix 
        ----------------------------------
        dim( H ) = ( dof , dof )
        
        such that --> Kinetic Energy = 0.5 * v^T * H(q) * v
        
        """  
        
        M = np.diag( np.ones( self.dof ) ) # Default is identity matrix
        
        return M
    
    
    ###########################################################################
    def C(self, q , v ):
        """ 
        Corriolis and Centrifugal Matrix 
        ------------------------------------
        dim( C ) = ( dof , dof )
        
        such that --> d H / dt =  C + C^T
        
        
        """ 
        
        C = np.zeros( ( self.dof , self.dof ) ) # Default is zeros matrix
        
        return C
    
    ###########################################################################
    def N(self, q ):
        """ 
        Transformation matrix from generalized velocities to derivatives of
        configuration variables
        ------------------------------------
        dim( N ) = ( pos , dof )

        dq = N(q) v
        ------------------------------------
        """
        
        N = np.zeros( ( self.pos , self.dof ) )
        
        for i in range(min( self.pos, self.dof) ):
            N[i,i] = 1                # Diag matrix for the first m rows
        
        return N
    
    
    ###########################################################################
    def B(self, q ):
        """ 
        Actuator Matrix  : dof x m
        """
        
        B = np.zeros( ( self.dof , self.m ) )
        
        for i in range(min(self.m,self.dof)):
            B[i,i] = 1                # Diag matrix for the first m rows
        
        return B
    
    
    ###########################################################################
    def g(self, q ):
        """ 
        Gravitationnal forces vector : dof x 1
        """
        
        g = np.zeros( self.dof ) # Default is zero vector
        
        return g
    
        
    ###########################################################################
    def d(self, q , v ):
        """ 
        State-dependent dissipative forces : dof x 1
        """
        
        d = np.zeros(self.dof ) # Default is zero vector
        
        return d
    
    
    ###########################################################################
    # No need to overwrite the following functions for custom system
    ###########################################################################
    
    #############################
    def x2q( self, x ):
        """ from state vector (x) to angle and speeds (q,dq) """
        
        q  = x[ 0        : self.pos ]
        v  = x[ self.pos : self.n   ]
        
        return [ q , v ]
        
        
    #############################
    def q2x( self, q , v ):
        """ from angle and speeds (q,dq) to state vector (x) """
        
        x = np.zeros( self.n )
        
        x[ 0        : self.pos ] = q
        x[ self.pos : self.n   ] = v
        
        return x
    
    
    #############################
    def xut2q( self, x , u , t ):
        """ compute configuration variables """
        
        return self.x2q(x)[0]
    
    
    ##############################
    def generalized_forces(self, q  , v  , dv , t = 0 ):
        """ Computed generalized forces given a trajectory """  
        
        M = self.M( q )
        C = self.C( q , v )
        g = self.g( q )
        d = self.d( q , v )
                
        # Generalized forces
        forces = M @ dv + C @ v + g + d
        
        return forces
    
    
    ##############################
    def actuator_forces(self, q  , v  , dv , t = 0 ):
        """ Computed actuator forces given a trajectory (inverse dynamic) """  
        
        if self.dof == self.m:
        
            B = self.B( q )
                    
            # Generalized forces
            forces = self.generalized_forces( q , v , dv , t )
            
            # Actuator forces
            u = np.linalg.inv( B ) @ forces 
            
            return u
        
        else:
            
            raise NotImplementedError
        
    
    ##############################
    def accelerations(self, q , v , u , t = 0 ):
        """ 
        Compute accelerations vector (foward dynamic) 
        given :
        - actuator forces 
        - actual position and velocities
        """  
        
        M = self.M( q )
        C = self.C( q , v )
        g = self.g( q  )
        d = self.d( q , v)
        B = self.B( q )
        
        dv = np.linalg.inv( M ) @ ( B @ u - C @ v - g - d )
        
        return dv
    
    
    ###########################################################################
    def f(self, x , u , t = 0 ):
        """ 
        Continuous time foward dynamics evaluation
        
        dx = f(x,u,t)
        
        INPUTS
        x  : state vector             n x 1
        u  : control inputs vector    m x 1
        t  : time                     1 x 1
        
        OUPUTS
        dx : state derivative vectror n x 1
        
        """
        
        # from state vector (x) to position and velocities 
        [ q , v ] = self.x2q( x )       
        
        # compute accelerations
        dv = self.accelerations( q , v , u , t )

        # compute derivative of position varibles
        dq = self.N( q ) @ v
        
        # convert to state vector diff (dx)
        dx = self.q2x( dq , dv )        
        
        return dx
    
    
    ###########################################################################
    def kinetic_energy(self, q  , v ):
        """ Compute kinetic energy  """  

        e = 0.5 * v.T @ self.M( q ) @ v
        
        return e





###############################################################################
        
class GeneralizedMechanicalSystemWithPositionInputs( GeneralizedMechanicalSystem ):
    """ 
    Generalized Mechanical system with position inputs

    -------------------------------------------------------
    M(q) dv + C(q,v) v + d(q,v,u) + g(q) = B(q,u) e(u)
    dq = N(q) v
    -------------------------------------------------------

    numper of inputs are:
    m = m_f + m_o
    ---------------------------------------------------
    m      :   integer         : number of inputs
    m_f    :   integer         : number of force inputs
    m_o    :   integer         : number of other inputs
    u      :   dim = ( m , 1)  : vector of all input variables

    v        :  dim = (dof, 1)   : velocity variables
    q        :  dim = (pos, 1)   : position variables 
    dq       :  dim = (pos, 1)   : derivatives of position variables  
    dv       :  dim = (dof, 1)   : acceleration variables
    e        :  dim = (m_f, 1)   : force input variables
    d(q,v,u) :  dim = (dof, 1)   : state-dependent dissipative forces
    g(q)     :  dim = (dof, 1)   : state-dependent conservatives forces
    M(q)     :  dim = (dof, dof) : inertia matrix
    C(q,v)   :  dim = (dof, dof) : corriolis matrix
    B(q,u)   :  dim = (dof, m_f) : actuator matrix
    N(q)     :  dim = (pos, dof) : transformation matrix
    
    """
    
    ############################
    def __init__(self, dof = 1 , force_inputs = 1, other_inputs = 1, pos = None):
        """ """
        
        # Degree of Freedom
        self.dof = dof

        self.m_f = force_inputs
        self.m_o = other_inputs

        # Nb of configurations
        if pos == None:   # If not specifyied 
            pos = dof
        self.pos = pos
        
        # Dimensions
        n = dof + pos
        m = self.m_f + self.m_o
        p = n
        
        # initialize standard params
        system.ContinuousDynamicSystem.__init__(self, n, m, p)
        
        # Name
        self.name = str(dof) + 'DoF Generalized Mechanical System'
        
        # Labels, bounds and units
        for i in range(pos):
            # positions states
            self.x_ub[i] = + 10
            self.x_lb[i] = - 10
            self.state_label[i] = 'Position '+ str(i)
            self.state_units[i] = '[m]'
        for i in range(dof):
            # generalized velocity states
            j = i + pos
            self.x_ub[j] = + 10
            self.x_lb[j] = - 10
            self.state_label[j] = 'Velocity ' + str(i)
            self.state_units[j] = '[m/sec]'
        for i in range(self.m_f):
            self.u_ub[i] = + 5
            self.u_lb[i] = - 5
            self.input_label[i] = 'Force input ' + str(i)
            self.input_units[i] ='[N]'
        self.output_label = self.state_label
        self.output_units = self.state_units
            
    ###########################################################################
    # The following functions needs to be overloaded by child classes
    # to represent the dynamic of the system
    ###########################################################################
        
    #############################
    def u2e( self, u ):
        """  
        extract force inputs from all inputs
        """
        
        e = u[ 0 : self.m_f ] 
        
        return e
    
    ###########################################################################
    def B(self, q , u ):
        """ 
        Actuator Matrix  : dof x m
        """
        
        B = np.zeros( ( self.dof , self.m_f ) )
        
        for i in range(min( self.m_f , self.dof )):
            B[i,i] = 1   # Diag matrix for the first m rows
        
        return B
    
        
    ###########################################################################
    def d(self, q , v , u ):
        """ 
        State-dependent dissipative forces : dof x 1
        """
        
        d = np.zeros(self.dof ) # Default is zero vector
        
        return d
    
    
    ###########################################################################
    # No need to overwrite the following functions for custom system
    ###########################################################################
    
    ##############################
    def generalized_forces(self, q  , v  , dv , t = 0 ):
        
        raise NotImplementedError
    
    
    ##############################
    def actuator_forces(self, q  , v  , dv , t = 0 ):
        """ Computed actuator forces given a trajectory (inverse dynamic) """  
        
        raise NotImplementedError
        
    
    ##############################
    def accelerations(self, q , v , u , t = 0 ):
        """ 
        Compute accelerations vector (foward dynamic) 
        given :
        - actuator forces 
        - actual position and velocities
        """  
        
        M = self.M( q )
        C = self.C( q , v )
        g = self.g( q  )
        d = self.d( q , v,  u )
        B = self.B( q , u )

        e = self.u2e( u )
        
        dv = np.linalg.inv( M ) @ ( B @ e - C @ v - g - d )
        
        return dv
    
    

###############################################################################
        
class RigidBody2D( GeneralizedMechanicalSystemWithPositionInputs ):    
    """

    Mechanical system with Equations of Motion written in 
    a body-fixed frame of reference

    """

    ############################
    def __init__(self, force_inputs = 2, other_inputs = 0):
        """ """
        
        # Degree of Freedom
        self.dof = 3
        self.pos = 3

        self.m_f = force_inputs
        self.m_o = other_inputs

        # Dimensions
        n = 6
        m = self.m_f + self.m_o
        p = n
        
        # initialize standard params
        system.ContinuousDynamicSystem.__init__(self, n, m, p)
        
        # Name
        self.name = 'Planar Rigid Body'
        self.state_label = ['x','y','theta','v1','v2','w']
        self.state_units = ['[m]','[m]','[rad]','[m/sec]','[m/sec]','[rad/sec]']
        self.input_label = ['u1','u2']
        self.input_units = ['[]','[]']
        self.output_label = self.state_label
        self.output_units = self.state_units

        # Dynamic properties
        self.mass     = 1.0
        self.inertia  = 1.0
        self.l_t      = 1.0 # Distance between CG and Thrust

        # Default graphic output parameters
        self.dynamic_domain  = True


    ###########################################################################
    def M(self, q ):
        
        M = np.diag( np.array([ self.mass , self.mass, self.inertia ]) )
        
        return M
    
    
    ###########################################################################
    def C(self, q , v ):

        C = np.zeros( ( self.dof , self.dof ) ) 
        
        w = v[2]
        
        C[1,0] = + self.mass * w
        C[0,1] = - self.mass * w
        
        return C
    
    ###########################################################################
    def N(self, q ):
        """ 
        Transformation matrix from generalized velocities to derivatives of
        configuration variables
        ------------------------------------
        dim( N ) = ( pos , dof )

        dq = N(q) v
        ------------------------------------
        """

        theta = q[2]

        N = np.array( [ [ np.cos(theta) , -np.sin(theta)  , 0 ] ,
                        [ np.sin(theta) , +np.cos(theta)  , 0 ] ,
                        [ 0             , 0               , 1 ] ] )
        
        return N
    
    ###########################################################################
    def g(self, q ):
        """ 
        Gravitationnal forces vector : dof x 1
        """
        
        g = np.zeros( self.dof ) # Default is zero vector
        
        return g
    
        
    ###########################################################################
    def d(self, q , v , u):
        """ 
        State-dependent dissipative forces : dof x 1
        """
        
        d = np.zeros(self.dof ) # Default is zero vector
        
        return d
    
    ###########################################################################
    def B(self, q , u ):
        """ 
        Actuator Matrix 
        ------------------
        This placeholder is for a 2D point force [ F_x , F_y ] 
        applied at a point located at a distance l_t behind the CG
        """
        
        B = np.zeros(( self.dof, self.m_f ))
        
        B[0,0] = 1
        B[1,1] = 1
        B[2,1] = - self.l_t 
        
        return B
    
    ###########################################################################
    def forward_kinematic_domain(self, q ):
        """
        Place holder graphical output ( box with a force )
        """

        l = self.l_t * 10
        
        x = q[0]
        y = q[1]
        z = 0
        
        if self.dynamic_domain:
        
            domain  = [ ( -l + x , l + x ) ,
                        ( -l + y , l + y ) ,
                        ( -l + z , l + z ) ]#  
        else:
            
            domain  = [ ( -l , l ) ,
                        ( -l , l ) ,
                        ( -l , l ) ]#
                
        return domain
    

    ###########################################################################
    def forward_kinematic_lines(self, q ):
        """
        Place holder graphical output ( box with a force )
        """
        
        lines_pts = [] # list of array (n_pts x 3) for each lines
        lines_style = []
        lines_color = []
        
        ###########################
        #  body
        ###########################
        
        x     = q[0]
        y     = q[1]
        theta = q[2]
        
        W_T_B    = geometry.transformation_matrix_2D( theta , x , y )

        w = self.l_t

        # Points in body frame
        pts      = np.zeros(( 5 , 3 ))
        pts[0,:] = np.array([-w,+w,0])
        pts[1,:] = np.array([-w,-w,0])
        pts[2,:] = np.array([+w,-w,0])
        pts[3,:] = np.array([+w,+w,0])
        pts[4,:] = np.array([-w,+w,0])
        
        pts_W    = drawing.transform_points_2D( W_T_B , pts )

        lines_pts.append( pts_W )
        lines_style.append( '-')
        lines_color.append( 'b' )
        
        ###########################
        #  C.G.
        ###########################
        
        pts      = np.zeros(( 1 , 3 ))
        pts[0,:] = np.array([x,y,0])
        
        lines_pts.append( pts )
        lines_style.append( 'o')
        lines_color.append( 'b' )

        ###########################
        #  Diff flat output
        ###########################

        J = self.inertia
        m = self.mass
        r = self.l_t

        xp = x + J / (m * r) * np.cos(theta)
        yp = y + J / (m * r) * np.sin(theta)

        pts      = np.zeros(( 1 , 3 ))
        pts[0,:] = np.array([xp,yp,0])
        
        lines_pts.append( pts )
        lines_style.append( 'o')
        lines_color.append( 'r' )
                
        return lines_pts , lines_style , lines_color
    
    ###########################################################################
    def forward_kinematic_lines_plus(self, x , u , t ):
        """
        Place holder graphical output ( box with a force )
        """
        
        lines_pts = [] # list of array (n_pts x 3) for each lines
        lines_style = []
        lines_color = []

        # M per Newton of force
        f2r = 1.0 / self.u_ub[0] * self.l_t
        
        ###########################
        # force vector
        ###########################
        
        vx  = u[0] * f2r
        vy  = u[1] * f2r
        offset = -self.l_t
        
        pts_body = drawing.arrow_from_components( vx , vy , x = offset, origin = 'tip'  )    
        W_T_B    = geometry.transformation_matrix_2D( x[2], x[0] , x[1] )
        pts_W    = drawing.transform_points_2D( W_T_B , pts_body )
        
        lines_pts.append( pts_W )
        lines_style.append( '-')
        lines_color.append( 'r' )

        return lines_pts , lines_style , lines_color

    
'''
#################################################################
##################          Main                         ########
#################################################################
'''


if __name__ == "__main__":     
    """ MAIN TEST """
    
    #sys = GeneralizedMechanicalSystem( dof = 2 , pos = 2 , actuators = 2 )
    #sys = GeneralizedMechanicalSystemWithPositionInputs( dof = 3 , pos = 1 , force_inputs= 1 , other_inputs=1 )
    sys = RigidBody2D()
    
    #sys.show(  q = np.array([ 1.0, 2.0, 0.5 ]) )
    
    sys.ubar = np.array([1,0.2])
    sys.x0   = np.array([0,0,0,0,0,0])
    
    sys.compute_trajectory( tf = 20 )
    sys.plot_trajectory()
    sys.animate_simulation()
        


# -*- coding: utf-8 -*-
"""
Created on Jun 2 2021

@author: Alex
"""
import numpy as np

from pyro.dynamic               import massspringdamper
from pyro.control.linear        import PIDController


# Plant
sys = massspringdamper.ThreeMass()

sys.m1 = 2
sys.m2 = 3
sys.k1 = 0.01
sys.k2 = 2
sys.b1 = 0.1
sys.b2 = 0.2

sys.compute_ABCD()

sys.C = np.array([[0, 0, 1.0, 0, 0, 0]])

sys.plot_linearized_bode()

# PID controller
kp  = 10
ki  = 1
kd  = 1

tau = 0.05

ctl = PIDController(kp, ki, kd, tau)

ctl.rbar[0] = 2 # reference

# Simulation Closed-Loop Non-linear with LQR controller
cl_sys = ctl + sys
cl_sys.x0[2] = 0
cl_sys.compute_trajectory( 300 )
cl_sys.plot_trajectory('xu')
cl_sys.plot_linearized_pz_map()
cl_sys.plot_linearized_bode()
cl_sys.animate_simulation()


# -*- coding: utf-8 -*-
"""
Created on Mon Oct 22 11:37:48 2018

@author: alxgr
"""
###############################################################################
import numpy as np
###############################################################################
from pyro.control import controller
###############################################################################

###############################################################################
# Function
###############################################################################

def to_2D_arr(arr):
    arr = np.asanyarray(arr)
    if arr.ndim == 2:
        return arr

    if arr.ndim == 1:
        return arr[np.newaxis]
    elif arr.ndim == 0:
        return arr[np.newaxis, np.newaxis]
    else:
        raise ValueError(
            "Cannot expand array with %d dimensions to 2-D" % (arr.ndim)
        )



###############################################################################
# Simple proportionnal controller
###############################################################################
        
class ProportionnalSingleVariableController( controller.StaticController ) :
    """ 
    Simple proportionnal compensator
    ---------------------------------------
    r  : reference signal vector       k x 1
    y  : sensor signal vector          k x 1
    u  : control inputs vector         k x 1
    t  : time                          1 x 1
    ---------------------------------------
    u = c( y , r , t ) = (r - y) * gain

    """
    
    ###########################################################################
    # The two following functions needs to be implemented by child classes
    ###########################################################################
    
    
    ############################
    def __init__(self, k = 1):
        """ """
        
        # Dimensions
        self.k = k   
        self.m = k   
        self.p = k
        
        controller.StaticController.__init__( self, self.k, self.m, self.p)
        
        # Label
        self.name = 'Proportionnal Controller'
        
        # Gains
        self.gain = 1
        
    
    #############################
    def c( self , y , r , t = 0 ):
        """ 
        Feedback static computation u = c(y,r,t)
        
        INPUTS
        y  : sensor signal_proc vector     p x 1
        r  : reference signal_proc vector  k x 1
        t  : time                     1 x 1
        
        OUPUTS
        u  : control inputs vector    m x 1
        
        """
        
        u = np.zeros(self.m) # State derivative vector
        
        e = r - y
        
        u = self.gain * e
        
        return u


###############################################################################
# MIMO prop controller
###############################################################################
        
class ProportionalController(controller.StaticController):
    """
    General (SISO or MIMO) proportional controller
    -------------------------------------------------
    
    u = K * ( r - y )
    
    -----------------------------------------
    r  : reference signal vector       k x 1
    y  : sensor signal vector          p x 1
    u  : control inputs vector         m x 1
    -----------------------------------------
    
    """
    
    ###############################
    def __init__(self, m = 1 , p = 1 ):
        
        self.K = np.zeros((m,p))
        
        #self.K = to_2D_arr(K)
        #m = self.K.shape[0]
        #p = self.K.shape[1]
        
        controller.StaticController.__init__( self, p, m, p)
        
        self.rbar = np.zeros((self.k,))
        self.name = "%d X %d Proportional Controller" % self.K.shape
        
        self.ybar = np.zeros((self.p,))  # feedback offset
        self.ubar = np.zeros((self.m,))  # control input offset
        
    ############################
    @classmethod
    def from_matrix(cls, K):
        """  """
        
        K = to_2D_arr(K)
        
        m = K.shape[0]
        p = K.shape[1]
        
        instance = cls( m , p )
        
        instance.K = K
        
        return instance
        
        
    ##############################
    def c(self, y, r, t=0):
        """ Feedback law """
        
        delta_y = y - self.ybar
        delta_u = self.K.dot(r - delta_y)
        
        u = delta_u + self.ubar
        
        return u



###############################################################################
# MIMO PID controller
###############################################################################
class PIDController( controller.DynamicController ):
    """General (SISO or MIMO) PID controller

    Parameters
    ----------
    KP : array_like
        *m x p* Matrix of proportional controller gain
    KI : array_like
        *m x p* Matrix of integral controller gain
    KD : array_like
        *m x p* Matrix of derivative controller gain
    tau : float, optional
        Time constant of derivative filter.
    sat : float, optional
        Saturation of u signal

    Notes
    -----
    The error derivative is filtered and computed according to governing equations
    from:
    https://www.mathworks.com/help/physmod/sps/ref/filteredderivativediscreteorcontinuous.html

    """
    
    ##########################################################
    def __init__(self, KP, KI=None, KD=None, tau=3E-3 , sat = None):
        
        self.KP = to_2D_arr(KP)

        if KI is None:
            self.KI = np.zeros(self.KP.shape)
        else:
            self.KI = to_2D_arr(KI)
            if self.KI.shape != self.KP.shape:
                raise ValueError("Shape of KI does not match KP")

        if KD is None:
            self.KD = np.zeros(self.KP.shape)
        else:
            self.KD = to_2D_arr(KD)
            if self.KD.shape != self.KP.shape:
                raise ValueError("Shape of KD does not match KP")

        self.tau = tau
        self.sat = sat # saturation
        
        
        k = self.KP.shape[1]
        l = self.KP.shape[1]*2
        m = self.KP.shape[0]
        p = self.KP.shape[1]

        controller.DynamicController.__init__( self, k, l, m, p)
        
        for i in range(p):
            self.internal_state_label[i] = 'Integral of output ' + str(i)
            self.internal_state_label[i+self.p] = ('Filter state of output ' 
                                                   + str(i) )
        self.name = "PID Controller"
        
        
    #################################
    def b(self, z, y, r, t):
        """ Evaluate derivative of controller state """

        if z.shape != (self.l,):
            raise ValueError("Expected z with shape (%d,)" % self.l)
        if y.shape != (self.p,) or r.shape != (self.p,):
            raise ValueError("Expected r and y with shape (%d,)" % self.p)

        # Error
        e = r - y

        # Integrator state derivative
        dz_integral = e

        # Filtered derivative state
        z_filter  = self.get_z_filter( z )
        dz_filter = (e - z_filter) / self.tau

        dz = np.concatenate([dz_integral, dz_filter], axis=0)
        assert dz.shape == (self.l,)
        
        return dz
    
    
    #################################
    def c(self, z, y, r, t):
        if z.shape != (self.l, ):
            raise ValueError("expected z with shape (%d,)" % self.l)

        # Instantaneous error
        e = r - y

        # Error integral value
        ei = self.get_z_integral( z )

        # Error derivative value
        de = (e - self.get_z_filter(z)) / self.tau
        
        # Control Input
        u = self.KP.dot( e ) + self.KI.dot( ei ) + self.KD.dot( de )
        
        # Saturation
        if self.sat is not None:
            u = np.clip( u , -self.sat , self.sat)

        return u
    
    
    #################################
    def get_z_integral(self, z):
        """ get intergral error internal states """
        
        return z[:self.p]
    
    
    #################################
    def get_z_filter(self, z):
        """ get filter internal states """
        
        return z[self.p:]


'''
#################################################################
##################          Main                         ########
#################################################################
'''


if __name__ == "__main__":
    from pyro.dynamic import integrator
    
    plant = integrator.DoubleIntegrator()
    
    ctl      = PIDController( 10 , 0.1 , 5)
    ctl.rbar = np.array([ 2 ])
    
    sys = ctl + plant
    
    
    sys.compute_trajectory()
    sys.plot_trajectory('xu')
    sys.plot_trajectory_with_internal_states('x')
    
    
    
    
    
    
    


# -*- coding: utf-8 -*-
"""
Created on Tue Nov 13 11:05:07 2018

@author: Alexandre
"""

###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic import vehicle_steering
from pyro.planning import randomtree
###############################################################################

sys  = vehicle_steering.KinematicCarModelwithObstacles()

###############################################################################

# Planning
planner = randomtree.RRT( sys )

planner.x_start = np.array([0,0,0])
planner.x_goal  = np.array([30,0,0])

# Set domain
sys.x_ub = np.array([+35,+3.5,+3])
sys.x_lb = np.array([-5,-2,-3])

# Control actions
speed    = 2
steering = 0.3

planner.u_options = [
        np.array([ speed, +steering]),
        np.array([ speed, -steering]),
        np.array([ speed, 0]),
        np.array([-speed, 0])
        ]

planner.goal_radius = 1.0
planner.dt          = 0.1
planner.steps       = 5
planner.max_nodes   = 10000
planner.max_distance_compute = 50000

planner.compute_solution()

planner.plot_tree()
planner.plot_tree_3d()
planner.show_solution()
planner.animate_solution()





# -*- coding: utf-8 -*-
"""
Created on Mon Oct 22 08:40:31 2018

@author: alxgr
"""

import numpy as np
import matplotlib.pyplot as plt

from pyro.dynamic import system

from pyro.analysis import phaseanalysis
from pyro.analysis import simulation
from pyro.analysis import graphical
from pyro.analysis import costfunction

###############################################################################
# Mother Controller class
###############################################################################

class StaticController():
    """ 
    Mother class for memoryless controllers
    ---------------------------------------
    r  : reference signal vector       k x 1
    y  : sensor signal vector          p x 1
    u  : control inputs vector         m x 1
    t  : time                          1 x 1
    ---------------------------------------
    u = c( y , r , t )
    
    """
    
    ###########################################################################
    # The two following functions needs to be implemented by child classes
    ###########################################################################
    
    
    ############################
    def __init__(self, k=1, m=1, p=1):
        """ """
        
        # Dimensions
        self.k = k   
        self.m = m   
        self.p = p
        
        # Label
        self.name = 'Static Controller'
        
        # Reference signal info
        self.ref_label = []
        self.ref_units = []
        
        for i in range(k):
            self.ref_label.append('Ref. '+str(i))
            self.ref_units.append('')
        
        self.r_ub = np.zeros(self.k) + 10 # upper bounds
        self.r_lb = np.zeros(self.k) - 10 # lower bounds
        
        # default constant reference
        self.rbar = np.zeros(self.k)
        
    
    #############################
    def c( self , y , r , t = 0 ):
        """ 
        Feedback static computation u = c( y, r, t)
        
        INPUTS
        y  : sensor signal vector          p x 1
        r  : reference signal vector       k x 1
        t  : time                          1 x 1
        
        OUTPUTS
        u  : control inputs vector         m x 1
        
        """
        
        u = np.zeros(self.m) # State derivative vector
        
        raise NotImplementedError
        
        return u
    
    #########################################################################
    # Default methods that can be overloaded in child classes
    #########################################################################
    
    #############################
    def t2r( self , t ):
        """ 
        Reference signal fonction u = t2u(t)
        
        INPUTS
        t  : time                     1 x 1
        
        OUTPUTS
        r  : controller reference vector    m x 1
        
        Defaul is a constant signal equal to self.rbar, can overload the
        with a more complexe reference signal time-function 
        
        """
        
        #Default is a constant signal
        r = self.rbar
        
        return r
    
    #########################################################################
    def forward_kinematic_lines_plus( self, x , u , t ):
        """  
        Graphical output for the controller
        -----------------------------------
        default is nothing

        x,u,t are the state, input and time of the global closed-loop system

        """

        pts = None
        style = None
        color = None

        return pts, style, color
    
    
    #########################################################################
    # No need to overwrite the following functions for child classes
    #########################################################################
    
    #############################
    def cbar( self , y , t = 0 ):
        """ 
        Feedback static computation u = c( y, r = rbar, t) for
        default reference
        
        INPUTS
        y  : sensor signal vector     p x 1
        t  : time                     1 x 1
        
        OUTPUTS
        u  : control inputs vector    m x 1
        
        """
        r = self.t2r( t )
        u = self.c( y , r , t )
        
        return u
    
    
    #############################
    def __add__(self, sys):
        """ 
        closed_loop_system = controller + dynamic_system
        """
        
        cl_sys = ClosedLoopSystem( sys , self )
        
        return cl_sys
    
    
    #############################
    def plot_control_law(self, i=0, j=1, k=0, t=0, n = 10, sys = None):
        """ 
        k = control input index to plot
        i = state to use as the x-axis
        j = state to use as the y-axis
        n = grid resolution
        sys can be passed for state label unit and range
        """
        
        # Extract sys info
        
        if sys is not None:
            xname = sys.state_label[i] + ' ' + sys.state_units[i]
            yname = sys.state_label[j] + ' ' + sys.state_units[j]
            xmax  = sys.x_ub[i]
            xmin  = sys.x_lb[i]
            ymax  = sys.x_ub[j]
            ymin  = sys.x_lb[j]
            xbar  = sys.xbar
            
        else:
            xname = 'state x[%i]'%i
            yname = 'state x[%i]'%j
            xmax  = 10
            xmin  = -10
            ymax  = 10
            ymin  = -10
            xbar  = np.zeros( self.p )
            
        # Computing
        
        x = np.linspace( xmin  , xmax  , n )
        y = np.linspace( ymin  , ymax  , n )
        
        X, Y = np.meshgrid( x, y)
        
        U = np.zeros((n,n)) # control action table
        
        for l in range(n):
            for m in range(n):
                
                # Actual states
                x  = np.copy( xbar )   # default value for all states
                x[ i ] = X[l, m]
                x[ j ] = Y[l, m]
                
                # Control action
                u = self.cbar( x , t ) 
                
                U[l, m] = u[k] # extract control input element k
                
        
        # Ploting
        fig = plt.figure(figsize=(3, 2),dpi=300, frameon=True)
        
        fig.canvas.manager.set_window_title('Control law for u[%i]'%i)
        ax  = fig.add_subplot(1,1,1)
        
        ax.tick_params('both',labelsize = 5 )
        plt.ylabel(yname, fontsize = 5 )
        plt.xlabel(xname, fontsize = 5 )
        
        im1 = plt.pcolormesh( X , Y , U, shading='gouraud' , cmap = 'bwr')
        
        cbar = plt.colorbar(im1)
        cbar.ax.tick_params(labelsize=5)
        
        
        plt.axis([xmin,xmax,ymin,ymax])

        plt.grid(True)
        plt.tight_layout() 
        plt.show()
    



###############################################################################
# Mother "Static controller + dynamic system" class
###############################################################################

class ClosedLoopSystem( system.ContinuousDynamicSystem ):
    """ 
    Dynamic system connected with a static controller
    ---------------------------------------------
    NOTE: 
    Ignore any feedthough in the plant to avoid creating algebraic loops
    This is only valid if the output function h is not a fonction of u
    New equations assume y = h(x,u,t) -- > y = h(x,t)

    """
    ############################
    def __init__(self, ContinuousDynamicSystem , StaticController ):
        """ """
        
        self.plant      = ContinuousDynamicSystem
        self.controller = StaticController
        
        ######################################################################
        # Check dimensions match
        if not (self.plant.m == self.controller.m ):
            raise NameError('Dimension mismatch between controller and' + 
            ' dynamic system for the input signal u')
        elif not (self.plant.p == self.controller.p ):
            raise NameError('Dimension mismatch between controller and' + 
            ' dynamic system for the output signal y')
        ######################################################################
        
        # Dimensions of global closed-loop dynamic system
        self.n = self.plant.n
        self.m = self.controller.k 
        self.p = self.plant.p
        
        # Labels
        self.name = ('Closed-Loop ' + self.plant.name + 
                     ' with ' + self.controller.name )
        self.state_label  = self.plant.state_label
        self.input_label  = self.controller.ref_label
        self.output_label = self.plant.output_label
        
        # Units
        self.state_units  = self.plant.state_units
        self.input_units  = self.controller.ref_units
        self.output_units = self.plant.output_units
        
        # Define the domain
        self.x_ub = self.plant.x_ub
        self.x_lb = self.plant.x_lb
        self.u_ub = self.controller.r_ub
        self.u_lb = self.controller.r_lb
        
        # Plot params
        self.domain           = self.plant.domain
        self.linestyle        = self.plant.linestyle
        self.linestyle_plus   = self.plant.linestyle_plus
        self.linescolor       = self.plant.linescolor
        self.linescolor_plus  = self.plant.linescolor_plus
        self.lines_plus       = self.plant.lines_plus
        self.is_3d            = self.plant.is_3d
        
        # Default State and inputs        
        self.xbar = self.plant.xbar
        self.tbar = self.plant.tbar
        self.ubar = self.controller.rbar
        
        ################################
        # Variables
        ################################
        
        # Initial value for simulations
        self.x0   = self.plant.x0
        
        # Result of last simulation
        self.traj = None
        
        # Cost function for evaluation
        # default is a quadratic cost function with diag Q and R matrices
        self.cost_function = costfunction.QuadraticCostFunction.from_sys(self)
        
    
    ###########################################################################
    def f( self , x , u , t ):
        """ 
        Continuous time foward dynamics evaluation dx = f(x,u,t)
        
        INPUTS
        x  : state vector             n x 1
        u  : control inputs vector    m x 1
        t  : time                     1 x 1
        
        OUTPUTS
        dx : state derivative vector  n x 1
        
        """
        
        dx = np.zeros(self.n) # State derivative vector
        
        r = u # input of closed-loop global sys is ref of the controller
        
        # Compute output signal
        y = self.plant.h( x, self.plant.ubar, t)
        
        # Compute control inputs
        u = self.controller.c( y, r, t)
        
        # Compute state derivatives
        dx = self.plant.f( x, u, t)
        
        return dx
    

    ###########################################################################
    def h( self , x , u , t ):
        """ 
        Output fonction y = h(x,u,t)
        
        INPUTS
        x  : state vector             n x 1
        u  : control inputs vector    m x 1
        t  : time                     1 x 1
        
        OUTPUTS
        y  : output derivative vector o x 1
        
        """
        
        #y = np.zeros(self.p) # Output vector
        
        # Using u = ubar to avoid algeabric loops
        
        y = self.plant.h( x , self.plant.ubar , t )
        
        return y
    
    
    ###########################################################################
    def t2u( self , t ):
        """ 
        Reference signal fonction u = t2u(t)
        
        INPUTS
        t  : time                     1 x 1
        
        OUTPUTS
        u  : control inputs vector    m x 1
        
        Defaul is a constant signal equal to self.ubar, can overload the
        with a more complexe reference signal time-function 
        
        """
        
        # Input of closed-loop global sys is ref of the controller
        u = self.controller.t2r(t)
        
        return u
    

    ###########################################################################
    # Place holder graphical output, overload with specific graph output
    ###########################################################################
        
    #############################
    def xut2q( self, x , u , t ):
        """ Compute configuration variables ( q vector ) """
        
        # Use the plant function
        q = self.plant.xut2q( x, u, t)
        
        return q
    
    
    ###########################################################################
    def forward_kinematic_domain(self, q ):
        """ Set the domain range for ploting, can be static or dynamic """

        # Use the plant function
        domain = self.plant.forward_kinematic_domain( q )
        
        return domain
    
    
    ###########################################################################
    def forward_kinematic_lines(self, q ):
        """ 
        Compute points p = [x;y;z] positions given config q 
        ----------------------------------------------------
        - points of interest for ploting
        
        Outpus:
        lines_pts = [] : a list of array (n_pts x 3) for each lines
        
        """

        lines_pts = self.plant.forward_kinematic_lines( q )
                
        return lines_pts
    
    
    ###########################################################################
    def forward_kinematic_lines_plus(self, x , u , t ):
        """  
        Return combined graphical output for the controller and the system
        """

        # TODO: this is a quick fix, need to be improved

        sys = self.plant.forward_kinematic_lines_plus( x , u , t )
        ctl = self.controller.forward_kinematic_lines_plus( x , u , t )

        if type(sys) is tuple:
            lines_pts   = sys[0]
            lines_style = sys[1]
            lines_color = sys[2]
        else:
            # Legacy graph function to remove eventually
            lines_pts   = sys
            lines_style = []
            lines_color = []
            for j, line in enumerate(lines_pts):
                lines_style.append( self.plant.linestyle  )  # default value 
                lines_color.append( self.plant.linescolor )  # default value 
        
        if ctl[0] is not None:
            lines_pts    = ctl[0] + lines_pts
            lines_style  = ctl[1] + lines_style
            lines_color  = ctl[2] + lines_color
                
        return lines_pts, lines_style, lines_color
    
    #############################################################################
    #### Updated shortcuts
    #############################################################################
    
    
    ###########################################################################
    def plot_phase_plane_closed_loop(self , x_axis = 0 , y_axis = 1 ):
        """ 
        Plot Phase Plane vector field of the system
        ------------------------------------------------
        
        blue arrows for the open-loop behavior
        red arrows  for the closed-loop behavior
        
        """

        pp = phaseanalysis.PhasePlot( self , x_axis , y_axis )
        
        pp.compute_grid()
        pp.plot_init()
        
        # Closed-loop Behavior
        pp.color = 'r'
        pp.compute_vector_field()
        pp.plot_vector_field()
        
        # Open-Loop Behavior
        pp.f     = self.plant.f
        pp.ubar  = self.plant.ubar
        pp.color = 'b'
        pp.compute_vector_field()
        pp.plot_vector_field()
        
        pp.plot_finish()
        
        pp.phasefig.show()
        
        return pp
        
    
    #############################
    def compute_trajectory(
        self, tf=10, n=10001, solver='solve_ivt'):
        """ 
        Simulation of time evolution of the system
        ------------------------------------------------
        tf : final time
        n  : time steps
        """

        sim = simulation.CLosedLoopSimulator(self, tf, n, solver)
        
        self.traj = sim.compute()

        return self.traj


    #############################################
    # Make graph function use the internal sys
    #############################################
    
    
    ###########################################################################
    def get_plotter(self):
        return self.plant.get_plotter()
    
    
    ###########################################################################
    def get_animator(self):

        return graphical.Animator(self)
    

    ###########################################################################
    def show(self, q , x_axis = 0 , y_axis = 1 ):
        """ Plot figure of configuration q """
        
        system.ContinuousDynamicSystem.show( self.plant , q , 
                                            x_axis = 0 , y_axis = 1  )
        
    
    ###########################################################################
    def show3(self, q ):
        """ Plot figure of configuration q """
        
        system.ContinuousDynamicSystem.show3( self.plant, q )
        
        
    ###########################################################################
    def plot_phase_plane_trajectory(self, x_axis=0, y_axis=1):
        """
        Plot a trajectory in the Phase Plane
        ---------------------------------------------------------------
        note: will call compute_trajectory if no simulation data is present
        
        """
        
        # Check is trajectory is already computed
        if self.traj == None:
            self.compute_trajectory()
            
        traj = self.traj
        
        pp = phaseanalysis.PhasePlot( self , x_axis , y_axis )
        pp.plot()

        plt.plot(traj.x[:,x_axis], traj.x[:,y_axis], 'b-') # path
        plt.plot([traj.x[0,x_axis]], [traj.x[0,y_axis]], 'ko') # start
        plt.plot([traj.x[-1,x_axis]], [traj.x[-1,y_axis]], 'rx') # end
        
        plt.draw()

        pp.phasefig.tight_layout()
        
        plt.draw()
        plt.show()
        
        
    ###########################################################################
    def plot_phase_plane_trajectory_closed_loop(self, x_axis=0, y_axis=1):
        """ 
        Plot Phase Plane vector field of the system and the trajectory
        ------------------------------------------------
        
        blue arrows for the open-loop behavior
        red arrows  for the closed-loop behavior
        
        
        """
        
        pp = phaseanalysis.PhasePlot( self , x_axis , y_axis )
        
        pp.compute_grid()
        pp.plot_init()
        
        # Closed-loop Behavior
        pp.color = 'r'
        pp.compute_vector_field()
        pp.plot_vector_field()
        
        # Open-Loop Behavior
        pp.f     = self.plant.f
        pp.ubar  = self.plant.ubar
        pp.color = 'b'
        pp.compute_vector_field()
        pp.plot_vector_field()
        
        # Check is trajectory is already computed
        if self.traj == None:
            self.compute_trajectory()
            
        traj = self.traj
        
        plt.plot(traj.x[:,x_axis], traj.x[:,y_axis], 'b-') # path
        plt.plot([traj.x[0,x_axis]], [traj.x[0,y_axis]], 'ko') # start
        plt.plot([traj.x[-1,x_axis]], [traj.x[-1,y_axis]], 'rx') # end
        
        pp.plot_finish()
        
        pp.phasefig.show()
        
        
        
    ###########################################################################
    def plot_end_effector_trajectory(self, traj = None ):
        
        self.plant.plot_end_effector_trajectory( self.traj )
        
        




###############################################################################
class DynamicController( StaticController ):
    """
    
    Mother class for controller with internal states and dynamics (memory)
    
    ex: integral action of a PID

    ----------------------------------------
    z  : controller internal states    l x 1
    r  : reference signal vector       k x 1
    y  : sensor signal vector          p x 1
    u  : control inputs vector         m x 1
    t  : time                          1 x 1
    -----------------------------------------
    
    Control law
    u  = c( z, y, r, t)
    
    Internal dynamic
    dz / dt  = b( z, y, r, t)
    
    """
    
    #############################
    def __init__(self, k, l, m, p):
        
        self.l = l
        self.m = m
        self.p = p
        self.k = k

        self.name = "Dynamic Controller"
        
        ############################
        # Reference signal info
        self.ref_label = []
        self.ref_units = []
        
        for i in range(k):
            self.ref_label.append('Ref. '+str(i))
            self.ref_units.append('')
        
        self.r_ub = np.zeros(self.k) + 10 # upper bounds
        self.r_lb = np.zeros(self.k) - 10 # lower bounds
        
        # default constant reference
        self.rbar = np.zeros(self.k)
        
        ###########################
        # Internal states info
        self.internal_state_label = []
        self.internal_state_units = []
        
        for i in range(l):
            self.internal_state_label.append('Internal state ' +str(i))
            self.internal_state_units.append('')
        
        self.z_ub = np.zeros(self.l) + 10 # upper bounds
        self.z_lb = np.zeros(self.l) - 10 # lower bounds
        
        # default constant reference
        self.zbar = np.zeros(self.l)
        
        # initial internal controller states
        self.z0   = np.zeros(self.l)
        
        

    #############################
    def c(self, z, y, r, t):
        """ 
        CONTROL LAW
        u = c( z, y, r, t)
        
        INPUTS
        z  : internal states               l x 1
        y  : sensor signal vector          p x 1
        r  : reference signal vector       k x 1
        t  : time                          1 x 1
        
        OUTPUTS
        u  : control inputs vector         m x 1
        
        """
        
        u = np.zeros( self.m ) 
        
        return u

    
    #############################
    def b(self, z, y, r, t):
        """ 
        INTERNAL CONTROLLER DYNAMIC
        dz/dt = b( z, y, r, t)
        
        INPUTS
        z  : internal states               l x 1
        y  : sensor signal vector          p x 1
        r  : reference signal vector       k x 1
        t  : time                          1 x 1
        
        OUTPUTS
        d z / dt  : time derivative of internal states        l x 1
        """
        
        dz = np.zeros( self.l )
        
        return dz
    
    
    #############################
    def cbar( self , y , t = 0 ):
        """ 
        Feedback static computation u = c( z = zbar, y, r = rbar, t) for
        default reference and internal states
        
        INPUTS
        y  : sensor signal vector     p x 1
        t  : time                     1 x 1
        
        OUTPUTS
        u  : control inputs vector    m x 1
        
        """
        
        u = self.c( self.zbar, y , self.rbar , t )
        
        return u
    

    #########################################################################
    def forward_kinematic_lines_plus( self, x, u , t ):
        """  
        Graphical output for the controller
        -----------------------------------
        default is nothing

        x,u,t are the state, input and time of the global closed-loop system

        """

        pts = None
        style = None
        color = None

        return pts, style, color
        
    
    #############################
    def __add__(self, sys):
        
        return DynamicClosedLoopSystem( sys, self)
    
        

##############################################################################
class DynamicClosedLoopSystem( ClosedLoopSystem ):
    """
    Closed loop system with Dynamic controller
    --------------------------------------------
    
    Global closed-loop system with physical plant states and virtual
    controller internal states
    
    x_global = [ x_plant ; z_controller ]
    
    """
    
    #######################################
    def __init__(self, plant, controller):
        
        # Check dimensions
        if plant.p != controller.p:
            raise ValueError("Controller inputs do not match system outputs")
        if plant.m != controller.m:
            raise ValueError("Controller outputs do not match system inputs")
            
            
        ########################
        #Remove cost funtion
        ########################
        
        plant.cost_function = None
        
        ClosedLoopSystem.__init__( self, plant, controller)

        # Add extra states that represent system memory
        self.n = self.plant.n + self.controller.l
        
        self.state_label = ( self.plant.state_label +
                             self.controller.internal_state_label )
        self.state_units = ( self.plant.state_units +
                             self.controller.internal_state_units )
        
        self.x_ub = np.concatenate([ self.plant.x_ub,
                                     self.controller.z_ub
                                    ], axis=0)
        self.x_lb = np.concatenate([ self.plant.x_lb,
                                     self.controller.z_lb
                                    ], axis=0)
        self.xbar = np.concatenate([ self.plant.xbar,
                                     self.controller.zbar
                                    ], axis=0)
        
        ################################
        # Variables
        ################################
        
        # Initial value for simulations
        self.x0   = np.concatenate([ self.plant.x0,
                                     self.controller.z0
                                    ], axis=0)
        
        # Result of last simulation
        self.traj = None
        
        # Cost function for evaluation
        # default is a quadratic cost function with diag Q and R matrices
        self.cost_function = None #costfunction.QuadraticCostFunction.from_sys(self)
        
            
    ######################################
    def f(self, x, u, t):
        """ 
        Continuous time foward dynamics evaluation dx = f(x,u,t)
        
        INPUTS
        x  : state vector             n x 1
        u  : control inputs vector    m x 1
        t  : time                     1 x 1
        
        OUTPUTS
        dx : state derivative vector  n x 1
        
        """
        
        x, z = self._split_states( x )

        # Input to global system interpreted as reference signal
        r = u

        # Eval current system output. Assume there is no feedforward term,
        # as it would cause an algebraic loop
        y = self.plant.h( x, self.plant.ubar, t)

        # input u to dynamic system evaluated by controller
        u = self.controller.c( z, y, r, t)
        
        # Time derivative of states
        dx = self.plant.f( x, u, t)
        dz = self.controller.b( z, y, r, t)

        dx = np.concatenate([ dx, dz], axis=0)
        assert dx.shape == (self.n,)
        
        return dx
    
    
    ######################################
    def fzbar(self, x_plant , u, t = 0):
        """ 
        Continuous time foward dynamics evaluation dx = f(x,u,t)
        
        with 
        z = zbar
        r = u
        
        INPUTS
        x  : state vector             plant.n x 1
        t  : time                     1 x 1
        
        OUTPUTS
        dx : state derivative vector  n x 1
        
        """
        
        # Input to global system interpreted as reference signal
        r = u

        # Eval current system output. Assume there is no feedforward term,
        # as it would cause an algebraic loop
        y = self.plant.h( x_plant, self.plant.ubar, t)

        # input u to dynamic system evaluated by controller
        u = self.controller.c( self.controller.zbar, y, r, t)
        
        # Time derivative of states
        dx = self.plant.f( x_plant, u, t)
        
        return dx
    
    
    ##########################################
    def h(self, x, u, t):
        """ 
        Output fonction y = h(x,u,t)
        
        INPUTS
        x  : state vector             n x 1
        u  : control inputs vector    m x 1
        t  : time                     1 x 1
        
        OUTPUTS
        y  : output derivative vector p x 1
        
        """
        
        x, z = self._split_states( x )
        
        y    = self.plant.h( x, u, t)
        
        return y
    
    
    #######################################
    def _split_states(self, x):
        """
        Separate full state vector into system and controller states
        
        """
        
        x_sys, x_ctl = x[:self.plant.n], x[self.plant.n:]
        assert x_ctl.shape == (self.controller.l,)
        
        return (x_sys, x_ctl)
    
    
    #############################
    def xut2q( self, x , u , t ):
        """ Compute configuration variables ( q vector ) """
        
        x , z = self._split_states( x )
        
        # Use the plant function
        q = self.plant.xut2q( x, u, t)
        
        return q
    
    
    #############################
    def compute_trajectory(
        self, tf=10, n=10001, solver='solve_ivt'):
        """ 
        Simulation of time evolution of the system
        ------------------------------------------------
        tf : final time
        n  : time steps
        """
        
        sim = simulation.DynamicCLosedLoopSimulator( self, tf, n, solver)
        
        self.traj = sim.compute()

        return self.traj
    
    
    #############################
    def plot_trajectory_with_internal_states(self, plot='x', **kwargs):
        """
        Plot time evolution of a simulation of this system
        ------------------------------------------------
        note: will call compute_trajectory if no simulation data is present

        """
        
        # Check if trajectory is already computed
        if self.traj == None:
            self.compute_trajectory()
            
        plotter = graphical.TrajectoryPlotter( self )
        plotter.plot( self.traj, plot, **kwargs)
        
    #############################
    def plot_internal_controller_states(self, plot='z', **kwargs):
        """
        Plot time evolution of a simulation of this system
        ------------------------------------------------
        note: will call compute_trajectory if no simulation data is present

        """
        
        # Check if trajectory is already computed
        if self.traj == None:
            self.compute_trajectory()
               
        plotter = graphical.TrajectoryPlotter( self )
        plotter.plot( self.traj, plot, **kwargs)


    
    ###########################################################################
    def plot_phase_plane_closed_loop( self , x_axis = 0 , y_axis = 1 ):
        """ 
        Plot Phase Plane vector field of the system
        ------------------------------------------------
        
        blue arrows for the open-loop behavior
        red arrows  for the closed-loop behavior
        
        """

        pp = phaseanalysis.PhasePlot( self.plant , x_axis , y_axis )
        
        pp.compute_grid()
        pp.plot_init()
        
        # Closed-loop Behavior
        pp.color = 'b'
        pp.compute_vector_field()
        pp.plot_vector_field()
        
        # Open-Loop Behavior
        pp.f     = self.fzbar    # assume default internal states
        pp.ubar  = self.ubar
        pp.color = 'r'
        pp.compute_vector_field()
        pp.plot_vector_field()
        
        pp.plot_finish()
        
        return pp


'''
#################################################################
##################          Main                         ########
#################################################################
'''


if __name__ == "__main__":     
    """ MAIN TEST """
    pass

    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Mar  1 14:08:03 2024

@author: alex
"""

import numpy as np
import matplotlib.pyplot as plt

import gymnasium as gym
from gymnasium import spaces


#################################################################
# Create a Gym Environment from a Pyro System
#################################################################
class Sys2Gym(gym.Env):
    """Create a Gym Environment from a Pyro System

    Taken from the Pyro system:
    - x0: nominal initial state
    - f: state evolution function xdot = f(x,u,t)
    - g: cost function g(x,u,t) (reward = -g)
    - h: observation function y = h(x,u,t)
    - x_ub, x_lb: state boundaries
    - u_ub, u_lb: control boundaries

    Additionnal parameters of the gym wrapper are:
    - dt: time step for the integration
    - tf: maximum duration of an episode
    - t0: initial time (only relevant if the system is time dependent)
    - render_mode: None or "human" for rendering the system
    - reset_mode: "uniform", "gaussian" or "determinist"
    - clipping_inputs: True if the system clips the control inputs
    - clipping_states: True if the system clips the states
    - x0_lb, x0_ub: boundaries for the initial state (only relevant if reset_mode is "uniform")
    - x0_std: standard deviation for the initial state (only relevant if reset_mode is "gaussian")
    - termination_is_reached: method to define the termination condition of the task (default is always False representing a continous time task)

    """

    metadata = {"render_modes": ["human"]}

    #################################################################
    def __init__(
        self, sys, dt=0.05, tf=10.0, t0=0.0, render_mode=None, reset_mode="uniform"
    ):

        # Boundaries
        self.t0 = t0
        self.tf = tf  # For truncation of episodes
        self.observation_space = spaces.Box(
            sys.x_lb, sys.x_ub
        )  # default is state feedback
        self.action_space = spaces.Box(sys.u_lb, sys.u_ub)

        # Dynamic evolution
        self.sys = sys
        self.dt = dt
        self.clipping_inputs = True  # The sys is assumed to clip out of bounds inputs
        self.clipping_states = False  # The sys is assumed to clip out of bounds states (some gym env do that but this is a very fake behavior not exibiited by real systems, to use with caution)

        # Rendering
        self.render_mode = render_mode

        # Reset parameters (stocasticity of the initial state)
        self.reset_mode = reset_mode
        # self.reset_mode = "uniform"
        self.x0_lb = sys.x0 + 0.1 * sys.x_lb
        self.x0_ub = sys.x0 + 0.1 * sys.x_ub
        # self.reset_mode = "gaussian"
        self.x0_std = 0.1 * (sys.x_ub - sys.x_lb)
        # self.reset_mode = "determinist"

        # Memory
        self.x = sys.x0
        self.u = sys.ubar
        self.t = t0

        # Init
        self.render_is_initiated = False

        if self.render_mode == "human":
            self.init_render()

    #################################################################
    def reset(self, seed=None, options=None):

        if self.reset_mode == "uniform":

            super().reset(seed=seed)

            self.x = self.np_random.uniform(self.x0_lb, self.x0_ub)
            self.u = self.sys.ubar
            self.t = self.t0

        elif self.reset_mode == "gaussian":

            super().reset(seed=seed)

            self.x = self.np_random.normal(self.sys.x0, self.x0_std)
            self.u = self.sys.ubar
            self.t = 0.0

        else:
            # Deterministic
            self.x = self.sys.x0
            self.u = self.sys.ubar
            self.t = 0.0

        # Observation
        y = self.sys.h(self.x, self.u, self.t)

        # Info
        info = {"state": self.x, "action": self.u}

        return y, info

    #################################################################
    def step(self, u):

        # Clipping of inputs
        if self.clipping_inputs:
            u = np.clip(u, self.sys.u_lb, self.sys.u_ub)

        # State and time at the beginning of the step
        x = self.x
        t = self.t
        dt = self.dt

        # Derivatives
        dx = self.sys.f(x, u, t)

        # Euler integration #TODO use a better integrator
        x_new = x + dx * dt
        t_new = t + dt

        # Horrible optionnal hack to avoid the system to go out of bounds
        if self.clipping_states:
            x_new = np.clip(x_new, self.sys.x_lb, self.sys.x_ub)

        # Termination of episodes
        terminated = self.termination_is_reached()

        # Reward = negative of cost function
        if terminated:
            r = -self.sys.cost_funtion.h(x, t)  # Terminal cost
        else:
            r = (
                -self.sys.cost_function.g(x, u, t) * dt
            )  # Instantaneous cost integrated over the time step

        # Truncation of episodes if out of space-time bounds
        truncated = (t_new > self.tf) or (not self.sys.isavalidstate(x_new))

        # Memory update
        self.x = x_new
        self.t = t_new
        self.u = u  # The memory of the control input is only used for redering

        # Observation
        y = self.sys.h(self.x, self.u, self.t)

        # Info
        info = {"state": self.x, "action": self.u}

        # Rendering
        if self.render_mode == "human":
            self.render()

        return y, r, terminated, truncated, info

    #################################################################
    def init_render(self):

        self.render_is_initiated = True

        self.animator = self.sys.get_animator()
        self.animator.show_plus(self.x, self.u, self.t)
        plt.pause(0.001)

    #################################################################
    def render(self):

        if self.render_mode == "human":
            if not self.render_is_initiated:
                self.init_render()
            self.animator.show_plus_update(self.x, self.u, self.t)
            plt.pause(0.001)

    #################################################################
    def termination_is_reached(self):
        """This method should be overloaded in the child class to define the termination condition for a task that is not time defined in continous time."""

        return False


"""
#################################################################
##################          Main                         ########
#################################################################
"""


if __name__ == "__main__":
    """MAIN TEST"""

    from pyro.dynamic import pendulum

    from stable_baselines3 import PPO

    sys = pendulum.InvertedPendulum()

    # Physical parameters
    sys.gravity = 10.0
    sys.m1 = 1.0
    sys.l1 = 1.0
    sys.lc1 = 0.5 * sys.l1
    sys.I1 = (1.0 / 12.0) * sys.m1 * sys.l1**2

    sys.l_domain = 2 * sys.l1  # graphical domain

    # Min/max state and control inputs
    sys.x_ub = np.array([+3.0 * np.pi, +8])
    sys.x_lb = np.array([-3.0 * np.pi, -8])
    sys.u_ub = np.array([+2.0])
    sys.u_lb = np.array([-2.0])

    # Time constant
    dt = 0.05

    # Cost Function
    # The reward function is defined as: r = -(theta2 + 0.1 * theta_dt2 + 0.001 * torque2)
    sys.cost_function.xbar = np.array([0, 0])  # target
    sys.cost_function.R[0, 0] = 0.001 / dt
    sys.cost_function.Q[0, 0] = 1.0 / dt
    sys.cost_function.Q[1, 1] = 0.1 / dt

    sys.x0 = np.array([-np.pi, 0.0])

    gym_env = Sys2Gym(sys, dt=dt, render_mode=None)

    gym_env.clipping_states = True # To reproduce the behavior of gym pendulum

    gym_env.reset_mode = "uniform"
    gym_env.x0_lb = np.array([-np.pi , -1.0])
    gym_env.x0_ub = np.array([+np.pi , +1.0])

    model = PPO("MlpPolicy", gym_env, verbose=1)
    model.learn(total_timesteps=250000)

    gym_env = Sys2Gym(sys, render_mode="human")

    #gym_env.reset_mode = "uniform"

    episodes = 3
    for episode in range(episodes):
        y, info = gym_env.reset()
        terminated = False
        truncated = False

        print("\n Episode:", episode)
        while not (terminated or truncated):
            u, _states = model.predict(y, deterministic=True)
            y, r, terminated, truncated, info = gym_env.step(u)


    # from pyro.control.reinforcementlearning import stable_baseline3_controller

    # ppo_ctl = stable_baseline3_controller(model)


    # ppo_ctl.plot_control_law(sys=sys, n=100)
    # cl_sys = ppo_ctl + sys

    # cl_sys.x0 = np.array([-3.0, -0.0])
    # cl_sys.compute_trajectory(tf=10.0, n=10000, solver="euler")
    # cl_sys.plot_trajectory("xu")
    # cl_sys.animate_simulation()
